import torch
import torch.nn as nn
import torch.nn.functional as F
from .utils import bilinear_sampler
from .setrans import CrossAttFeatTrans, gen_all_indices, SETransInputFeatEncoder
import os


class CorrBlock:
    def __init__(self, fmap1, fmap2, num_levels=4, radius=4, do_corr_global_norm=False):
        self.num_levels = num_levels
        self.radius = radius
        self.corr_pyramid = []
        self.do_corr_global_norm = do_corr_global_norm

        # all pairs correlation
        corr = CorrBlock.corr(fmap1, fmap2)

        batch, h1, w1, dim, h2, w2 = corr.shape
        if self.do_corr_global_norm:
            corr_3d = corr.permute(0, 3, 1, 2, 4, 5).view(batch, dim, -1)
            corr_normed = F.layer_norm(corr_3d, (corr_3d.shape[2],), eps=1e-12)
            corr = corr_normed.view(batch, dim, h1, w1, h2, w2).permute(
                0, 2, 3, 1, 4, 5
            )

        corr = corr.reshape(batch * h1 * w1, dim, h2, w2)

        # Save corr for visualization
        if "SAVECORR" in os.environ:
            corr_savepath = os.environ["SAVECORR"]
            # corr2: batch, dim, h1, w1, h2, w2.
            corr2 = (
                corr.reshape(batch, h1, w1, dim, h2, w2)
                .permute(0, 3, 1, 2, 4, 5)
                .detach()
                .cpu()
            )
            torch.save(corr2, corr_savepath)
            print(f"Corr tensor saved to {corr_savepath}")

        self.corr_pyramid.append(corr)
        for i in range(self.num_levels - 1):
            corr = F.avg_pool2d(corr, 2, stride=2)
            self.corr_pyramid.append(corr)

    def __call__(self, coords):
        r = self.radius
        coords = coords.permute(0, 2, 3, 1)
        batch, h1, w1, _ = coords.shape

        out_pyramid = []
        for i in range(self.num_levels):
            corr = self.corr_pyramid[i]
            dx = torch.linspace(
                -r, r, 2 * r + 1, dtype=coords.dtype, device=coords.device
            )
            dy = torch.linspace(
                -r, r, 2 * r + 1, dtype=coords.dtype, device=coords.device
            )
            delta = torch.stack(torch.meshgrid(dy, dx, indexing="ij"), axis=-1)

            centroid_lvl = coords.reshape(batch * h1 * w1, 1, 1, 2) / 2**i
            delta_lvl = delta.view(1, 2 * r + 1, 2 * r + 1, 2)
            coords_lvl = centroid_lvl + delta_lvl

            corr = bilinear_sampler(corr, coords_lvl)
            corr = corr.view(batch, h1, w1, -1)
            out_pyramid.append(corr)

        # Concatenate the four levels (4 resolutions of neighbors),
        # and permute the neighbors to the channel dimension.
        out = torch.cat(out_pyramid, dim=-1)
        # [batch, number of neighbors, h1, w1]
        return out.permute(0, 3, 1, 2).contiguous()

    @staticmethod
    def corr(fmap1, fmap2):
        batch, dim, ht, wd = fmap1.shape
        fmap1 = fmap1.view(batch, dim, ht * wd)
        fmap2 = fmap2.view(batch, dim, ht * wd)

        corr = torch.matmul(fmap1.transpose(1, 2), fmap2)
        corr = corr.view(batch, ht, wd, 1, ht, wd)
        return corr / torch.sqrt(torch.tensor(dim))


class CorrBlockSingleScale(nn.Module):
    def __init__(self, fmap1, fmap2, num_levels=4, radius=4, do_corr_global_norm=False):
        super().__init__()
        self.radius = radius
        self.do_corr_global_norm = do_corr_global_norm

        # all pairs correlation
        corr = CorrBlock.corr(fmap1, fmap2)
        batch, h1, w1, dim, h2, w2 = corr.shape
        if self.do_corr_global_norm:
            corr_3d = corr.permute(0, 3, 1, 2, 4, 5).view(B, dim, -1)
            corr_normed = F.layer_norm(corr_3d, (corr_3d.shape[2],), eps=1e-12)
            corr = corr_normed.view(batch, dim, h1, w1, h2, w2).permute(
                0, 2, 3, 1, 4, 5
            )

        self.corr = corr.reshape(batch * h1 * w1, dim, h2, w2)
        self.do_corr_global_norm = do_corr_global_norm

    def __call__(self, coords):
        r = self.radius
        coords = coords.permute(0, 2, 3, 1)
        batch, h1, w1, _ = coords.shape

        corr = self.corr
        dx = torch.linspace(-r, r, 2 * r + 1)
        dy = torch.linspace(-r, r, 2 * r + 1)
        delta = torch.stack(torch.meshgrid(dy, dx, indexing="ij"), axis=-1).to(
            coords.device
        )

        centroid_lvl = coords.reshape(batch * h1 * w1, 1, 1, 2)
        delta_lvl = delta.view(1, 2 * r + 1, 2 * r + 1, 2)
        coords_lvl = centroid_lvl + delta_lvl

        corr = bilinear_sampler(corr, coords_lvl)
        out = corr.view(batch, h1, w1, -1)
        out = out.permute(0, 3, 1, 2).contiguous()
        return out

    @staticmethod
    def corr(fmap1, fmap2):
        batch, dim, ht, wd = fmap1.shape
        fmap1 = fmap1.view(batch, dim, ht * wd)
        fmap2 = fmap2.view(batch, dim, ht * wd)

        corr = torch.matmul(fmap1.transpose(1, 2), fmap2)
        corr = corr.view(batch, ht, wd, 1, ht, wd)
        return corr / torch.sqrt(torch.tensor(dim))


# TransCorrBlock instance is created and destroyed in each call of raft.forward().
# It is only for a particular pair of image features fmap1, fmap2
class TransCorrBlock(CorrBlock, nn.Module):
    def __init__(self, config, num_levels=4, radius=4, do_corr_global_norm=False):
        # Do not call CorrBlock.__init__(), as corr is computed differently.
        nn.Module.__init__(self)
        self.num_levels = num_levels
        self.radius = radius
        self.config = config
        self.setrans = CrossAttFeatTrans(self.config, "Inter-frame correlation block")
        self.vispos_encoder = SETransInputFeatEncoder(self.config)
        self.coords2 = None
        self.do_corr_global_norm = do_corr_global_norm

    # Compute the correlation volume between all pairs here.
    # If fmap1o, fmap2o are None, then fmap1 is conv features, fmap2 is transformer features.
    # If fmap1o, fmap2o are not None, then fmap1, fmap2 are transformer features,
    # and fmap1o, fmap2o are conv features.
    def update(self, fmap1, fmap2, fmap1o, fmap2o, coords1, coords2=None):
        self.corr_pyramid = []
        # coords1 is generated by coords_grid(), with the format
        #           (width index,  height index)
        # flip  =>  (height index, width index)
        coords1 = coords1.permute(0, 2, 3, 1).flip(-1)
        if coords2 is None:
            coords2 = gen_all_indices(fmap2.shape[2:], device=fmap2.device)
            coords2 = coords2.unsqueeze(0).repeat(fmap2.shape[0], 1, 1, 1)

        vispos1, pos_biases = self.vispos_encoder(
            fmap1, coords1, return_pos_biases=True
        )
        vispos2 = self.vispos_encoder(fmap2, coords2, return_pos_biases=False)
        batch, dim, ht, wd = fmap1.shape

        # If both f1_trans and f2_trans are used, then compute the two-way correlation.
        # Otherwise, only fmap2o is not None, and compute the single-way correlation.
        if fmap1o is not None and fmap2o is not None:
            vispos1o = self.vispos_encoder(fmap1o, coords1, return_pos_biases=False)
            vispos2o = self.vispos_encoder(fmap2o, coords2, return_pos_biases=False)
            # Two-way correlation.
            corr_1t2o = self.corr(ht, wd, vispos1, vispos2o, pos_biases)
            corr_1o2t = self.corr(ht, wd, vispos1o, vispos2, pos_biases)
            # Try concatenation instead of averaging, as averaging may lose info.
            corr = torch.cat([corr_1t2o, corr_1o2t], dim=3)
        else:
            # single-way correlation.
            corr = self.corr(ht, wd, vispos1, vispos2, pos_biases)

        batch, h1, w1, dim, h2, w2 = corr.shape
        # Merge batch with h1 and w1 to improve efficiency. They will be separate later.
        corr = corr.reshape(batch * h1 * w1, dim, h2, w2)

        if "SAVECORR" in os.environ:
            corr_savepath = os.environ["SAVECORR"]
            corr2 = corr.detach().cpu().reshape(batch, h1, w1, h2, w2)
            torch.save(corr2, corr_savepath)
            print(f"Corr tensor saved to {corr_savepath}")

        self.corr_pyramid.append(corr)
        for i in range(self.num_levels - 1):
            corr = F.avg_pool2d(corr, 2, stride=2)
            self.corr_pyramid.append(corr)

    def corr(self, ht, wd, vispos1, vispos2, pos_biases):
        batch, ht_wd, dim = vispos1.shape
        assert ht_wd == ht * wd
        # if out_attn_only, output attention matrix is in the shape of (query unit number, key unit number)
        # otherwise, output features are in the same shape as the query features.
        # key features are recombined to get new query features by matmul(attention_probs, V(key features))
        #             frame1 frame2
        # corr: [1, 1, 7040, 7040]
        corr = self.setrans(vispos1, vispos2, pos_biases)
        if self.do_corr_global_norm:
            B, C, H, W = corr.shape
            corr_3d = corr.view(B, C, H * W)
            corr_normed = F.layer_norm(corr_3d, (corr_3d.shape[2],), eps=1e-12)
            corr = corr_normed.view(B, C, H, W)

        corr = corr.view(batch, ht, wd, 1, ht, wd)
        return corr

    # __call__() inherits from CorrBlock.



import torch


def build_flowformer(cfg):
    name = cfg.transformer
    if name == "percostformer3":
        from .PerCostFormer3.transformer import FlowFormer
    else:
        raise ValueError(f"FlowFormer = {name} is not a valid optimizer!")

    return FlowFormer(cfg[name])



from typing import Optional

from einops import rearrange
import torch
import torch.nn as nn
import torch.nn.functional as F

from ptlflow.utils.registry import register_model, trainable
from .corr import CorrBlock
from .gma import Attention
from .twins_csc import Twins_CSC
from .update import SKUpdateBlock_TAM_v3
from .utils import coords_grid
from ..base_model.base_model import BaseModel


class SequenceLoss(nn.Module):
    def __init__(self, gamma: float, max_flow: float):
        super().__init__()
        self.gamma = gamma
        self.max_flow = max_flow

    def forward(self, outputs, inputs):
        """Loss function defined over sequence of flow predictions"""

        flow_preds = outputs["flow_preds"]
        flow_gt = inputs["flows"][:, 0]
        valid = inputs["valids"][:, 0]

        n_predictions = len(flow_preds)
        flow_loss = 0.0

        # exlude invalid pixels and extremely large diplacements
        mag = torch.sum(flow_gt**2, dim=1, keepdim=True).sqrt()
        valid = (valid >= 0.5) & (mag < self.max_flow)

        for i in range(n_predictions):
            i_weight = self.gamma ** (n_predictions - i - 1)
            i_loss = (flow_preds[i] - flow_gt).abs()
            flow_loss += i_weight * (valid * i_loss).mean()

        return flow_loss


class StreamFlow(BaseModel):
    pretrained_checkpoints = {
        "kitti": "https://github.com/hmorimitsu/ptlflow/releases/download/weights1/streamflow-kitti-eaafa6ed.ckpt",
        "sintel": "https://github.com/hmorimitsu/ptlflow/releases/download/weights1/streamflow-sintel-af557e5e.ckpt",
        "spring": "https://github.com/hmorimitsu/ptlflow/releases/download/weights1/streamflow-spring-092f8a17.ckpt",
        "things": "https://github.com/hmorimitsu/ptlflow/releases/download/weights1/streamflow-things-c640255a.ckpt",
    }

    def __init__(
        self,
        decoder_dim: int = 256,
        corr_levels: int = 4,
        corr_radius: int = 4,
        num_heads: int = 1,
        pcupdater_conv: list[int] = [1, 7],
        T: int = 4,
        k_conv: list[int] = [1, 15],
        use_gma: bool = True,
        iters: int = 15,
        twins_pretrained_ckpt: Optional[str] = None,
        gamma: float = 0.8,
        max_flow: float = 400,
        **kwargs,
    ) -> None:
        super().__init__(
            output_stride=8, loss_fn=SequenceLoss(gamma, max_flow), **kwargs
        )

        self.ratio = 8
        self.corr_radius = corr_radius
        self.iters = iters
        self.context_dim = cdim = decoder_dim // 2
        self.hidden_dim = decoder_dim // 2

        # feature network, context network, and update block
        self.fnet = Twins_CSC(pretrained_ckpt=twins_pretrained_ckpt)
        self.cnet = Twins_CSC(pretrained_ckpt=twins_pretrained_ckpt)
        self.update_block = SKUpdateBlock_TAM_v3(
            decoder_dim=decoder_dim,
            num_heads=num_heads,
            use_gma=use_gma,
            pcupdater_conv=pcupdater_conv,
            corr_levels=corr_levels,
            corr_radius=corr_radius,
            T=T,
            k_conv=k_conv,
        )
        if use_gma:
            self.att = Attention(
                dim=cdim,
                heads=num_heads,
                dim_head=cdim,
            )
        else:
            self.att = None

    def freeze_untemporal(self):
        for name, param in self.named_parameters():
            if "temporal" in name:
                param.requires_grad = True
            else:
                param.requires_grad = False

    def unfreeze_parameters(self):
        for name, param in self.named_parameters():
            param.requires_grad = True

    def freeze_bn(self):
        for m in self.modules():
            if isinstance(m, nn.BatchNorm2d):
                m.eval()

    def initialize_flow(self, img, ratio=8):
        """Flow is represented as difference between two coordinate grids flow = coords1 - coords0"""
        N, C, H, W = img.shape
        coords0 = coords_grid(
            N, H // ratio, W // ratio, dtype=img.dtype, device=img.device
        )

        # optical flow computed as difference: flow = coords1 - coords0
        return coords0

    def upsample_flow(self, flow, mask, ratio=8):
        """Upsample flow field [H/8, W/8, 2] -> [H, W, 2] using convex combination"""
        N, _, H, W = flow.shape
        mask = mask.view(N, 1, 9, ratio, ratio, H, W)
        mask = torch.softmax(mask, dim=2)

        up_flow = F.unfold(ratio * flow, [3, 3], padding=1)
        up_flow = up_flow.view(N, 2, 9, 1, 1, H, W)

        up_flow = torch.sum(mask * up_flow, dim=2)
        up_flow = up_flow.permute(0, 1, 4, 2, 5, 3)
        return up_flow.reshape(N, 2, ratio * H, ratio * W)

    def forward(self, inputs):
        """Estimate optical flow between pair of frames"""
        images, image_resizer = self.preprocess_images(
            inputs["images"],
            bgr_add=-0.5,
            bgr_mult=2.0,
            bgr_to_rgb=True,
            resize_mode="pad",
            pad_mode="replicate",
            pad_two_side=True,
        )

        T = images.shape[1]

        hdim = self.hidden_dim
        cdim = self.context_dim

        # run the feature network
        fmaps = self.fnet(images)
        cnets = self.cnet(images[:, :-1])

        corr_fns = [
            CorrBlock(fmaps[:, i], fmaps[:, i + 1], radius=self.corr_radius)
            for i in range(T - 1)
        ]
        coord_0s = [
            self.initialize_flow(images[:, i], ratio=self.ratio) for i in range(T - 1)
        ]
        coord_1s = [
            self.initialize_flow(images[:, i], ratio=self.ratio) for i in range(T - 1)
        ]

        # if flow_init is not None:
        #     coord_1s = [coord_1s[i] + flow_init[i] for i in range(len(flow_init))]

        nets, inps, attentions = [], [], []
        nets, inps = torch.split(cnets, [hdim, cdim], dim=2)
        nets = torch.tanh(rearrange(nets, "B T C H W -> (B T) C H W"))
        inps = torch.relu(inps)
        inps = rearrange(inps, "B T C H W -> (B T) C H W")
        if self.att is not None:
            attentions = self.att(inps)
        else:
            attentions = None

        flow_predictions_list = [[] for i in range(T - 1)]
        for itr in range(self.iters):
            coord_1s = [coord.detach() for coord in coord_1s]
            corrs = rearrange(
                torch.stack([corr_fns[i](coord_1s[i]) for i in range(T - 1)], dim=1),
                "B T C H W -> (B T) C H W",
            )
            flows = rearrange(
                torch.stack([coord_1s[i] - coord_0s[i] for i in range(T - 1)], dim=1),
                "B T C H W -> (B T) C H W",
            )

            nets, up_masks, delta_flows = self.update_block(
                nets, inps, corrs, flows, attentions, T=T - 1
            )

            coord_1s = [coord_1s[i] + delta_flows[:, i] for i in range(T - 1)]
            for i in range(T - 1):
                flow_predictions_list[i].append(
                    self.upsample_flow(
                        coord_1s[i] - coord_0s[i], up_masks[:, i], ratio=self.ratio
                    )
                )

        out_flow = [
            self.postprocess_predictions(
                flow_predictions[-1], image_resizer, is_flow=True
            )
            for flow_predictions in flow_predictions_list
        ]
        out_flow = torch.stack(out_flow, 1)
        if self.training:
            outputs = {"flows": out_flow, "flow_preds": flow_predictions_list}
        else:
            flow_small = [coord_1s[i] - coord_0s[i] for i in range(T - 1)]
            flow_small = torch.stack(flow_small, 1)
            outputs = {"flows": out_flow, "flow_small": flow_small}

        return outputs


@register_model
@trainable
class streamflow(StreamFlow):
    pass



import torch
import torch.nn as nn
import timm
import numpy as np

# from .twins_ft import _twins_svt_large_jihao


class twins_svt_large(nn.Module):
    def __init__(self, pretrained=True, del_layers=True):
        super().__init__()
        self.svt = timm.create_model("twins_svt_large", pretrained=pretrained)

        if del_layers:
            del self.svt.head
            del self.svt.patch_embeds[2]
            del self.svt.patch_embeds[2]
            del self.svt.blocks[2]
            del self.svt.blocks[2]
            del self.svt.pos_block[2]
            del self.svt.pos_block[2]

    def forward(self, x, data=None, layer=2):
        B = x.shape[0]
        for i, (embed, drop, blocks, pos_blk) in enumerate(
            zip(
                self.svt.patch_embeds,
                self.svt.pos_drops,
                self.svt.blocks,
                self.svt.pos_block,
            )
        ):
            x, size = embed(x)
            x = drop(x)
            for j, blk in enumerate(blocks):
                x = blk(x, size)
                if j == 0:
                    x = pos_blk(x, size)
            if i < len(self.svt.depths) - 1:
                x = x.reshape(B, *size, -1).permute(0, 3, 1, 2).contiguous()

            if i == 0:
                x_16 = x.clone()
            if i == layer - 1:
                break

        return x

    def extract_ml_features(self, x, data=None, layer=2):
        res = []
        B = x.shape[0]
        for i, (embed, drop, blocks, pos_blk) in enumerate(
            zip(
                self.svt.patch_embeds,
                self.svt.pos_drops,
                self.svt.blocks,
                self.svt.pos_block,
            )
        ):
            x, size = embed(x)
            if i == layer - 1:
                x1 = x.reshape(B, *size, -1).permute(0, 3, 1, 2).contiguous()
            x = drop(x)
            for j, blk in enumerate(blocks):
                x = blk(x, size)
                if j == 0:
                    x = pos_blk(x, size)
            if i < len(self.svt.depths) - 1:
                x = x.reshape(B, *size, -1).permute(0, 3, 1, 2).contiguous()

            if i == layer - 1:
                break

        return x1, x

    def compute_params(self):
        num = 0

        for i, (embed, drop, blocks, pos_blk) in enumerate(
            zip(
                self.svt.patch_embeds,
                self.svt.pos_drops,
                self.svt.blocks,
                self.svt.pos_block,
            )
        ):
            for param in embed.parameters():
                num += np.prod(param.size())
            for param in blocks.parameters():
                num += np.prod(param.size())
            for param in pos_blk.parameters():
                num += np.prod(param.size())
            for param in drop.parameters():
                num += np.prod(param.size())
            if i == 1:
                break
        return num


class convnext_large(nn.Module):
    def __init__(self, pretrained=True):
        super().__init__()
        self.convnext = timm.create_model("convnext_large", pretrained=pretrained)

    def forward(self, x, data=None, layer=2):
        x = self.convnext.stem(x)
        x = self.convnext.stages[0](x)
        x = self.convnext.stages[1](x)
        return x

    def compute_params(self):
        num = 0

        for param in self.convnext.stem.parameters():
            num += np.prod(param.size())
        for param in self.convnext.stages[0].parameters():
            num += np.prod(param.size())
        for param in self.convnext.stages[1].parameters():
            num += np.prod(param.size())

        return num


class convnext_Xlarge_4x(nn.Module):
    def __init__(self, pretrained=True, del_layers=True):
        super().__init__()
        self.convnext = timm.create_model(
            "convnext_xlarge_in22k", pretrained=pretrained
        )

        # self.convnext.stem[0].stride = (2, 2)
        # self.convnext.stem[0].padding = (1, 1)

        if del_layers:
            del self.convnext.head
            del self.convnext.stages[1]
            del self.convnext.stages[1]
            del self.convnext.stages[1]

        # print(self.convnext)

    def forward(self, x, data=None, layer=2):
        x = self.convnext.stem(x)
        x = self.convnext.stages[0](x)
        return x


class convnext_base_2x(nn.Module):
    def __init__(self, pretrained=True, del_layers=True):
        super().__init__()
        self.convnext = timm.create_model("convnext_base_in22k", pretrained=pretrained)

        self.convnext.stem[0].stride = (2, 2)
        self.convnext.stem[0].padding = (1, 1)

        if del_layers:
            del self.convnext.head
            del self.convnext.stages[1]
            del self.convnext.stages[1]
            del self.convnext.stages[1]

        # print(self.convnext)

    def forward(self, x, data=None, layer=2):
        x = self.convnext.stem(x)
        x = self.convnext.stages[0](x)
        return x


if __name__ == "__main__":
    m = convnext_Xlarge_2x()
    input = torch.randn(2, 3, 64, 64)
    out = m(input)
    print(out.shape)



# =============================================================================
# Copyright 2024 Henrique Morimitsu
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# Adapted from IRR: https://github.com/visinf/irr/blob/master/models/pwc_modules.py
#
# Modifications by Henrique Morimitsu:
# - Remove operations not related to rescaling
# =============================================================================

import torch
import torch.nn.functional as F


def upsample2d_as(inputs, target_as, mode="bilinear"):
    _, _, h, w = target_as.size()
    if inputs.shape[-2] != h or inputs.shape[-1] != w:
        inputs = F.interpolate(inputs, [h, w], mode=mode, align_corners=True)
    return inputs


def rescale_flow(flow, width_im, height_im, to_local=True):
    if to_local:
        u_scale = float(flow.size(3) / width_im)
        v_scale = float(flow.size(2) / height_im)
    else:
        u_scale = float(width_im / flow.size(3))
        v_scale = float(height_im / flow.size(2))

    u, v = flow.chunk(2, dim=1)
    u = u * u_scale
    v = v * v_scale

    return torch.cat([u, v], dim=1)



import torch
import torch.nn as nn
import torch.nn.functional as F


class FlowHead(nn.Module):
    def __init__(
        self,
        input_dim=128,
        hidden_dim=256,
        out_dim=2,
    ):
        super(FlowHead, self).__init__()

        self.conv1 = nn.Conv2d(input_dim, hidden_dim, 3, padding=1)
        self.conv2 = nn.Conv2d(hidden_dim, out_dim, 3, padding=1)
        self.relu = nn.ReLU(inplace=True)

    def forward(self, x):
        out = self.conv2(self.relu(self.conv1(x)))

        return out


class SepConvGRU(nn.Module):
    def __init__(
        self,
        hidden_dim=128,
        input_dim=192 + 128,
        kernel_size=5,
    ):
        padding = (kernel_size - 1) // 2

        super(SepConvGRU, self).__init__()
        self.convz1 = nn.Conv2d(
            hidden_dim + input_dim, hidden_dim, (1, kernel_size), padding=(0, padding)
        )
        self.convr1 = nn.Conv2d(
            hidden_dim + input_dim, hidden_dim, (1, kernel_size), padding=(0, padding)
        )
        self.convq1 = nn.Conv2d(
            hidden_dim + input_dim, hidden_dim, (1, kernel_size), padding=(0, padding)
        )

        self.convz2 = nn.Conv2d(
            hidden_dim + input_dim, hidden_dim, (kernel_size, 1), padding=(padding, 0)
        )
        self.convr2 = nn.Conv2d(
            hidden_dim + input_dim, hidden_dim, (kernel_size, 1), padding=(padding, 0)
        )
        self.convq2 = nn.Conv2d(
            hidden_dim + input_dim, hidden_dim, (kernel_size, 1), padding=(padding, 0)
        )

    def forward(self, h, x):
        # horizontal
        hx = torch.cat([h, x], dim=1)
        z = torch.sigmoid(self.convz1(hx))
        r = torch.sigmoid(self.convr1(hx))
        q = torch.tanh(self.convq1(torch.cat([r * h, x], dim=1)))
        h = (1 - z) * h + z * q

        # vertical
        hx = torch.cat([h, x], dim=1)
        z = torch.sigmoid(self.convz2(hx))
        r = torch.sigmoid(self.convr2(hx))
        q = torch.tanh(self.convq2(torch.cat([r * h, x], dim=1)))
        h = (1 - z) * h + z * q

        return h


class BasicMotionEncoder(nn.Module):
    def __init__(
        self,
        corr_channels=324,
        flow_channels=2,
    ):
        super(BasicMotionEncoder, self).__init__()

        self.convc1 = nn.Conv2d(corr_channels, 256, 1, padding=0)
        self.convc2 = nn.Conv2d(256, 192, 3, padding=1)
        self.convf1 = nn.Conv2d(flow_channels, 128, 7, padding=3)
        self.convf2 = nn.Conv2d(128, 64, 3, padding=1)
        self.conv = nn.Conv2d(64 + 192, 128 - flow_channels, 3, padding=1)

    def forward(self, flow, corr):
        cor = F.relu(self.convc1(corr))
        cor = F.relu(self.convc2(cor))
        flo = F.relu(self.convf1(flow))
        flo = F.relu(self.convf2(flo))

        cor_flo = torch.cat([cor, flo], dim=1)
        out = F.relu(self.conv(cor_flo))
        return torch.cat([out, flow], dim=1)


class BasicUpdateBlock(nn.Module):
    def __init__(
        self,
        corr_channels=324,
        hidden_dim=128,
        context_dim=128,
        downsample_factor=8,
        flow_dim=2,
        bilinear_up=False,
    ):
        super(BasicUpdateBlock, self).__init__()

        self.encoder = BasicMotionEncoder(
            corr_channels=corr_channels,
            flow_channels=flow_dim,
        )

        self.gru = SepConvGRU(hidden_dim=hidden_dim, input_dim=context_dim + hidden_dim)

        self.flow_head = FlowHead(
            hidden_dim,
            hidden_dim=256,
            out_dim=flow_dim,
        )

        if bilinear_up:
            self.mask = None
        else:
            self.mask = nn.Sequential(
                nn.Conv2d(hidden_dim, 256, 3, padding=1),
                nn.ReLU(inplace=True),
                nn.Conv2d(256, downsample_factor**2 * 9, 1, padding=0),
            )

    def forward(self, net, inp, corr, flow):
        motion_features = self.encoder(flow, corr)

        inp = torch.cat([inp, motion_features], dim=1)

        net = self.gru(net, inp)
        delta_flow = self.flow_head(net)

        if self.mask is not None:
            mask = self.mask(net)
        else:
            mask = None

        return net, mask, delta_flow



import os
import math
import copy

import torch
import torch.nn as nn
from torch.nn import Parameter
import torch.nn.functional as F
from .setrans_ablation import (
    RandPosEmbedder,
    SinuPosEmbedder,
    ZeroEmbedder,
    MultiHeadFeatTrans,
)
from .utils import print0

torch.set_printoptions(sci_mode=False)

bb2_stage_dims = {
    "raft-small": [32, 32, 64, 96, 128],
    "raft-basic": [64, 64, 96, 128, 256],
    "resnet34": [64, 64, 128, 256, 512],
    "resnet50": [64, 256, 512, 1024, 2048],
    "resnet101": [64, 256, 512, 1024, 2048],
    "resibn101": [64, 256, 512, 1024, 2048],  # resibn: resnet + IBN layers
    "eff-b0": [16, 24, 40, 112, 1280],  # input: 224
    "eff-b1": [16, 24, 40, 112, 1280],  # input: 240
    "eff-b2": [16, 24, 48, 120, 1408],  # input: 260
    "eff-b3": [24, 32, 48, 136, 1536],  # input: 300
    "eff-b4": [24, 32, 56, 160, 1792],  # input: 380
    "i3d": [64, 192, 480, 832, 1024],  # input: 224
}


# Can also be implemented using torch.meshgrid().
def gen_all_indices(shape, device):
    indices = torch.arange(shape.numel(), device=device).view(shape)

    out = []
    for dim_size in reversed(shape):
        out.append(indices % dim_size)
        indices = torch.div(indices, dim_size, rounding_mode="trunc")
    return torch.stack(tuple(reversed(out)), len(shape))


# drop_path and DropPath are copied from timm/models/layers/drop.py
def drop_path(x, drop_prob: float = 0.0, training: bool = False):
    """Drop paths (Stochastic Depth) per sample (when applied in main path of residual blocks).

    This is the same as the DropConnect impl I created for EfficientNet, etc networks, however,
    the original name is misleading as 'Drop Connect' is a different form of dropout in a separate paper...
    See discussion: https://github.com/tensorflow/tpu/issues/494#issuecomment-532968956 ... I've opted for
    changing the layer and argument names to 'drop path' rather than mix DropConnect as a layer name and use
    'survival rate' as the argument.

    """
    if drop_prob == 0.0 or not training:
        return x
    keep_prob = 1 - drop_prob
    shape = (x.shape[0],) + (1,) * (
        x.ndim - 1
    )  # work with diff dim tensors, not just 2D ConvNets
    random_tensor = keep_prob + torch.rand(shape, dtype=x.dtype, device=x.device)
    random_tensor.floor_()  # binarize
    output = x.div(keep_prob) * random_tensor
    return output


class DropPath(nn.Module):
    """Drop paths (Stochastic Depth) per sample  (when applied in main path of residual blocks)."""

    def __init__(self, drop_prob=None):
        super(DropPath, self).__init__()
        self.drop_prob = drop_prob

    def forward(self, x):
        return drop_path(x, self.drop_prob, self.training)


class SETransConfig(object):
    def __init__(self):
        self.feat_dim = -1
        self.in_feat_dim = -1
        # self.backbone_type  = 'eff-b4'          # resnet50, resnet101, resibn101, eff-b1~b4
        # self.bb_stage_idx   = 4                 # Last stage of the five stages. Index starts from 0.
        # self.set_backbone_type(self.backbone_type)
        # self.use_pretrained = True

        # Positional encoding settings.
        self.pos_dim = 2
        self.pos_code_weight = 1

        # Architecture settings
        # Number of modes in the expansion attention block.
        # When doing ablation study of multi-head, num_modes means num_heads,
        # to avoid introducing extra config parameters.
        self.num_modes = 4
        self.tie_qk_scheme = "shared"  # shared, loose, or none.
        self.trans_output_type = "private"  # shared or private.
        self.act_fun = F.gelu

        self.attn_clip = 100
        self.attn_diag_cycles = 1000
        self.base_initializer_range = 0.02

        self.qk_have_bias = False
        # Without the bias term, V projection often performs better.
        self.v_has_bias = False
        # Add an identity matrix (*0.02*query_idbias_scale) to query/key weights
        # to make a bias towards identity mapping.
        # Set to 0 to disable the identity bias.
        self.query_idbias_scale = 10
        self.feattrans_lin1_idbias_scale = 10

        # Pooling settings
        self.pool_modes_feat = "softmax"  # softmax, max, mean, or none.

        # Randomness settings
        self.hidden_dropout_prob = 0.1
        self.attention_probs_dropout_prob = 0.2
        self.drop_path_prob = 0  # Drop path reduces performance greatly.
        self.pos_code_type = "bias"
        self.ablate_multihead = False
        self.out_attn_probs_only = False
        # When out_attn_scores_only, dropout is not applied to attention scores.
        self.out_attn_scores_only = False
        self.attn_mask_radius = -1

    def set_backbone_type(self, args):
        if self.try_assign(args, "backbone_type"):
            self.bb_stage_dims = bb2_stage_dims[self.backbone_type]
            self.in_feat_dim = self.bb_stage_dims[-1]

    # return True if any parameter is successfully set, and False if none is set.
    def try_assign(self, args, *keys):
        is_successful = False

        for key in keys:
            if key in args:
                if isinstance(args, dict):
                    self.__dict__[key] = args[key]
                else:
                    self.__dict__[key] = args.__dict__[key]
                is_successful = True

        return is_successful

    def update_config(self, args):
        self.set_backbone_type(args)
        self.try_assign(
            args,
            "use_pretrained",
            "apply_attn_stage",
            "num_modes",
            "trans_output_type",
            "base_initializer_range",
            "pos_code_type",
            "ablate_multihead",
            "attn_clip",
            "attn_diag_cycles",
            "tie_qk_scheme",
            "feattrans_lin1_idbias_scale",
            "qk_have_bias",
            "v_has_bias",
            # out_attn_probs_only/out_attn_scores_only are only True for the optical flow correlation block.
            "out_attn_probs_only",
            "out_attn_scores_only",
            "in_feat_dim",
            "pos_bias_radius",
        )

        if self.try_assign(args, "out_feat_dim"):
            self.feat_dim = self.out_feat_dim
        else:
            self.feat_dim = self.in_feat_dim

        if "dropout_prob" in args and args.dropout_prob >= 0:
            self.hidden_dropout_prob = args.dropout_prob
            self.attention_probs_dropout_prob = args.dropout_prob
            print0("Dropout prob: %.2f" % (args.dropout_prob))


CONFIG = SETransConfig()


# =================================== SETrans Initialization ====================================#
class SETransInitWeights(nn.Module):
    """An abstract class to handle weights initialization"""

    def __init__(self, config, *inputs, **kwargs):
        super(SETransInitWeights, self).__init__()
        self.config = config

    def init_weights(self, module):
        """Initialize the weights.
        type(module.weight)      # <class 'torch.nn.parameter.Parameter'>
        type(module.weight.data) # <class 'torch.Tensor'>
        """
        if isinstance(module, (nn.Linear, nn.Embedding)):
            base_initializer_range = self.config.base_initializer_range
            module.weight.data.normal_(mean=0.0, std=base_initializer_range)
            # Slightly different from the TF version which uses truncated_normal
            # for initialization cf https://github.com/pytorch/pytorch/pull/5617
        if isinstance(module, nn.Linear) and module.bias is not None:
            module.bias.data.zero_()


def tie_qk(module):
    if (
        isinstance(module, CrossAttFeatTrans)
        and module.tie_qk_scheme != "none"
        and module.tie_qk_scheme != None
    ):
        module.tie_qk()


def add_identity_bias(module):
    if isinstance(module, CrossAttFeatTrans) or isinstance(module, ExpandedFeatTrans):
        module.add_identity_bias()


# ====================================== SETrans Shared Modules ========================================#


class MMSharedMid(nn.Module):
    def __init__(self, config):
        super(MMSharedMid, self).__init__()
        self.num_modes = config.num_modes
        self.feat_dim = config.feat_dim
        self.shared_linear = nn.Linear(self.feat_dim, self.feat_dim)
        self.mid_act_fn = config.act_fun
        # This dropout is not presented in huggingface transformers.
        # Added to conform with lucidrains and rwightman's implementations.
        self.dropout = nn.Dropout(config.hidden_dropout_prob)

    # x: [B0, 1792*4, U]
    def forward(self, x):
        # shape_4d: [B0, 4, 1792, U].
        shape_4d = (x.shape[0], self.num_modes, self.feat_dim, x.shape[2])
        # x_4d: [B0, 4, U, 1792].
        x_4d = x.view(shape_4d).permute([0, 1, 3, 2])

        x_trans = self.shared_linear(x_4d)
        x_act = self.mid_act_fn(x_trans)
        x_drop = self.dropout(x_act)

        # restore the original shape
        x_drop = x_drop.permute([0, 1, 3, 2]).reshape(x.shape)

        return x_drop


# MMPrivateOutput/MMSharedOutput <- MMandedFeatTrans <- CrossAttFeatTrans
# MM***Output has a shortcut (residual) connection.
class MMPrivateOutput(nn.Module):
    def __init__(self, config):
        super(MMPrivateOutput, self).__init__()
        self.num_modes = config.num_modes
        self.feat_dim = config.feat_dim
        feat_dim_allmode = self.feat_dim * self.num_modes
        # Each group (mode) is applied a linear transformation, respectively.
        self.group_linear = nn.Conv1d(
            feat_dim_allmode, feat_dim_allmode, 1, groups=self.num_modes
        )
        self.resout_norm_layer = nn.LayerNorm(
            self.feat_dim, eps=1e-12, elementwise_affine=True
        )
        self.dropout = nn.Dropout(config.hidden_dropout_prob)

    # x, shortcut: [B0, 1792*4, U]
    def forward(self, x, shortcut):
        x = self.group_linear(x)
        # x_comb: [B0, 1792*4, U]. Residual connection.
        x_comb = x + shortcut
        shape_4d = (x.shape[0], self.num_modes, self.feat_dim, x.shape[2])
        # x_comb_4d, x_drop_4d: [B0, 4, U, 1792].
        x_comb_4d = x.view(shape_4d).permute([0, 1, 3, 2])
        x_drop_4d = self.dropout(x_comb_4d)
        x_normed = self.resout_norm_layer(x_drop_4d)
        return x_normed


# MMPrivateOutput/MMSharedOutput <- MMandedFeatTrans <- CrossAttFeatTrans
# MM***Output has a shortcut (residual) connection.
class MMSharedOutput(nn.Module):
    # feat_dim_allmode is not used. Just to keep the ctor arguments the same as MMPrivateOutput.
    def __init__(self, config):
        super(MMSharedOutput, self).__init__()
        self.num_modes = config.num_modes
        self.feat_dim = config.feat_dim
        self.shared_linear = nn.Linear(self.feat_dim, self.feat_dim)
        self.resout_norm_layer = nn.LayerNorm(
            self.feat_dim, eps=1e-12, elementwise_affine=True
        )
        self.dropout = nn.Dropout(config.hidden_dropout_prob)

    # x, shortcut: [B0, 1792*4, U] or [B0, 4, U, 1792]
    def forward(self, x, shortcut):
        # shape_4d: [B0, 4, 1792, U].
        shape_4d = (x.shape[0], self.num_modes, self.feat_dim, x.shape[2])
        if len(x.shape) == 3:
            x_4d = x.view(shape_4d).permute([0, 1, 3, 2])
        else:
            x_4d = x
        if len(shortcut.shape) == 3:
            shortcut_4d = shortcut.view(shape_4d).permute([0, 1, 3, 2])
        else:
            shortcut_4d = shortcut

        # x_4d, shortcut_4d: [B0, 4, U, 1792].
        x_trans = self.shared_linear(x_4d)
        # x_4d, x_comb: [B0, 4, U, 1792]. Residual connection.
        x_comb = x_trans + shortcut_4d
        x_drop = self.dropout(x_comb)
        x_normed = self.resout_norm_layer(x_drop)
        return x_normed


# group_dim: the tensor dimension that corresponds to the multiple groups.
class LearnedSoftAggregate(nn.Module):
    def __init__(self, num_feat, group_dim, keepdim=False):
        super(LearnedSoftAggregate, self).__init__()
        self.group_dim = group_dim
        # num_feat = 1: element-wise score function & softmax.
        # num_feat > 1: the linear score function is applied to the last dim (features) of the input tensor.
        self.num_feat = num_feat
        self.feat2score = nn.Linear(num_feat, 1)
        self.keepdim = keepdim

    def forward(self, x, score_basis=None):
        # Assume the last dim of x is the feature dim.
        if score_basis is None:
            score_basis = x

        if self.num_feat == 1:
            mode_scores = self.feat2score(score_basis.unsqueeze(-1)).squeeze(-1)
        else:
            mode_scores = self.feat2score(score_basis)
        attn_probs = mode_scores.softmax(dim=self.group_dim)
        x_aggr = (x * attn_probs).sum(dim=self.group_dim, keepdim=self.keepdim)
        return x_aggr


# ExpandedFeatTrans <- CrossAttFeatTrans.
# ExpandedFeatTrans has a residual connection.
class ExpandedFeatTrans(nn.Module):
    def __init__(self, config, name):
        super(ExpandedFeatTrans, self).__init__()
        self.config = config
        self.name = name
        self.in_feat_dim = config.in_feat_dim
        self.feat_dim = config.feat_dim
        self.num_modes = config.num_modes
        self.feat_dim_allmode = self.feat_dim * self.num_modes
        # first_linear is the value projection in other transformer implementations.
        # The output of first_linear will be divided into num_modes groups.
        # first_linear is always 'private' for each group, i.e.,
        # parameters are not shared (parameter sharing makes no sense).
        self.first_linear = nn.Linear(
            self.in_feat_dim, self.feat_dim_allmode, bias=config.v_has_bias
        )

        self.base_initializer_range = config.base_initializer_range
        self.has_FFN = getattr(config, "has_FFN", True)
        self.has_input_skip = getattr(config, "has_input_skip", False)
        self.drop_path = (
            DropPath(config.drop_path_prob)
            if config.drop_path_prob > 0.0
            else nn.Identity()
        )

        print0(
            "{}: v_has_bias: {}, has_FFN: {}, has_input_skip: {}".format(
                self.name, config.v_has_bias, self.has_FFN, self.has_input_skip
            )
        )

        self.pool_modes_keepdim = False
        self.pool_modes_feat = config.pool_modes_feat

        if self.pool_modes_feat == "softmax":
            agg_basis_feat_dim = self.feat_dim

            # group_dim = 1, i.e., features will be aggregated across the modes.
            self.feat_softaggr = LearnedSoftAggregate(
                agg_basis_feat_dim, group_dim=1, keepdim=self.pool_modes_keepdim
            )

        if self.has_FFN:
            self.intermediate = MMSharedMid(self.config)

            if config.trans_output_type == "shared":
                self.output = MMSharedOutput(config)
            elif config.trans_output_type == "private":
                self.output = MMPrivateOutput(config)

        # Have to ensure U1 == U2.
        if self.has_input_skip:
            self.input_skip_coeff = Parameter(torch.ones(1))
            self.skip_layer_norm = nn.LayerNorm(
                self.feat_dim, eps=1e-12, elementwise_affine=False
            )

    def add_identity_bias(self):
        if self.config.feattrans_lin1_idbias_scale > 0:
            # first_linear dimension is num_modes * feat_dim.
            # If in_feat_dim == feat_dim, only add identity bias to the first mode.
            # If in_feat_dim > feat_dim, expand to more modes until all in_feat_dim dimensions are covered.
            identity_weight = (
                torch.diag(torch.ones(self.feat_dim))
                * self.base_initializer_range
                * self.config.feattrans_lin1_idbias_scale
            )
            # Only bias the weight of the first mode.
            # The total initial "weight mass" in each row is reduced by 1792*0.02*0.5.
            self.first_linear.weight.data[: self.feat_dim, : self.feat_dim] = (
                self.first_linear.weight.data[: self.feat_dim, : self.feat_dim] * 0.5
                + identity_weight
            )

    # input_feat is usually key_feat.
    # input_feat: [3, 4416, 128]. attention_probs: [3, 4, 4416, 4416].
    def forward(self, input_feat, attention_probs):
        # input_feat: [B, U2, 1792], mm_first_feat: [B, Units, 1792*4]
        # B: batch size, U2: number of the 2nd group of units,
        # IF: in_feat_dim, could be different from feat_dim, due to layer compression
        # (different from squeezed attention).
        B, U2, IF = input_feat.shape
        U1 = attention_probs.shape[2]
        F = self.feat_dim
        M = self.num_modes
        mm_first_feat = self.first_linear(input_feat)
        # mm_first_feat after transpose: [B, 1792*4, U2]
        mm_first_feat = mm_first_feat.transpose(1, 2)

        # mm_first_feat_4d: [B, 4, U2, 1792]
        mm_first_feat_4d = mm_first_feat.view(B, M, F, U2).transpose(2, 3)

        # attention_probs:      [B, 4, U1, U2]. On sintel: [1, 4, 7040, 7040]
        # mm_first_feat_fusion: [B, 4, U2, F]. On sintel: [1, 4, 7040, 256]
        mm_first_feat_fusion = torch.matmul(attention_probs, mm_first_feat_4d)
        mm_first_feat_fusion_3d = mm_first_feat_fusion.transpose(2, 3).reshape(
            B, M * F, U1
        )
        mm_first_feat = mm_first_feat_fusion_3d

        if self.has_FFN:
            # mm_mid_feat:   [B, 1792*4, U1]. Group linear & gelu of mm_first_feat.
            mm_mid_feat = self.intermediate(mm_first_feat)
            # mm_last_feat:  [B, 4, U1, 1792]. Group/shared linear & residual & Layernorm
            mm_last_feat = self.output(mm_mid_feat, mm_first_feat)
            mm_trans_feat = mm_last_feat
        else:
            mm_trans_feat = mm_first_feat_fusion

        if self.pool_modes_feat == "softmax":
            trans_feat = self.feat_softaggr(mm_trans_feat)
        elif self.pool_modes_feat == "max":
            trans_feat = mm_trans_feat.max(dim=1)[0]
        elif self.pool_modes_feat == "mean":
            trans_feat = mm_trans_feat.mean(dim=1)
        elif self.pool_modes_feat == "none":
            trans_feat = mm_trans_feat

        # Have to ensure U1 == U2.
        if self.has_input_skip:
            trans_feat = self.input_skip_coeff * input_feat + self.drop_path(trans_feat)
            trans_feat = self.skip_layer_norm(trans_feat)

        # trans_feat: [B, U1, 1792]
        return trans_feat


class CrossAttFeatTrans(SETransInitWeights):
    def __init__(self, config, name):
        super(CrossAttFeatTrans, self).__init__(config)
        self.config = config
        self.name = name
        self.num_modes = config.num_modes
        self.in_feat_dim = config.in_feat_dim
        self.feat_dim = config.feat_dim
        self.attention_mode_dim = self.in_feat_dim // self.num_modes  # 448
        # att_size_allmode: 512 * modes
        self.att_size_allmode = self.num_modes * self.attention_mode_dim
        self.query = nn.Linear(
            self.in_feat_dim, self.att_size_allmode, bias=config.qk_have_bias
        )
        self.key = nn.Linear(
            self.in_feat_dim, self.att_size_allmode, bias=config.qk_have_bias
        )
        self.base_initializer_range = config.base_initializer_range

        self.out_attn_scores_only = config.out_attn_scores_only
        self.out_attn_probs_only = config.out_attn_probs_only
        self.ablate_multihead = config.ablate_multihead

        # out_attn_scores_only / out_attn_probs_only implies no FFN nor V projection.
        if self.out_attn_scores_only or self.out_attn_probs_only:
            self.out_trans = None
            if self.num_modes > 1:
                # Each attention value is a scalar. So num_feat = 1.
                self.attn_softaggr = LearnedSoftAggregate(1, group_dim=1, keepdim=True)

        elif self.ablate_multihead:
            self.out_trans = MultiHeadFeatTrans(config, name + "-out_trans")
        else:
            self.out_trans = ExpandedFeatTrans(config, name + "-out_trans")

        self.att_dropout = nn.Dropout(config.attention_probs_dropout_prob)

        self.tie_qk_scheme = config.tie_qk_scheme
        print0(
            "{}: in_feat_dim: {}, feat_dim: {}, modes: {}, qk_have_bias: {}".format(
                self.name,
                self.in_feat_dim,
                self.feat_dim,
                self.num_modes,
                config.qk_have_bias,
            )
        )

        # if using SlidingPosBiases2D, then add positional embeddings in CrossAttFeatTrans.forward().
        if config.pos_code_type == "bias":
            self.pos_code_weight = config.pos_code_weight
            print0("Positional biases weight: {:.3}".format(self.pos_code_weight))
        else:
            self.pos_code_weight = 1

        self.attn_clip = config.attn_clip
        if "attn_diag_cycles" in config.__dict__:
            self.attn_diag_cycles = config.attn_diag_cycles
        else:
            self.attn_diag_cycles = 1000
        self.max_attn = 0
        self.clamp_count = 0
        self.call_count = 0
        self.apply(self.init_weights)
        self.apply(tie_qk)
        # tie_qk() has to be executed after weight initialization.
        self.apply(add_identity_bias)

    # if tie_qk_scheme is not None, it overrides the initialized self.tie_qk_scheme
    def tie_qk(self, tie_qk_scheme=None):
        # override config.tie_qk_scheme
        if tie_qk_scheme is not None:
            self.tie_qk_scheme = tie_qk_scheme

        if self.tie_qk_scheme == "shared":
            self.key.weight = self.query.weight
            if self.key.bias is not None:
                self.key.bias = self.query.bias

        elif self.tie_qk_scheme == "loose":
            self.key.weight.data.copy_(self.query.weight)
            if self.key.bias is not None:
                self.key.bias.data.copy_(self.query.bias)

    def add_identity_bias(self):
        identity_weight = (
            torch.diag(torch.ones(self.attention_mode_dim))
            * self.base_initializer_range
            * self.config.query_idbias_scale
        )
        repeat_count = self.in_feat_dim // self.attention_mode_dim
        identity_weight = identity_weight.repeat([1, repeat_count])
        # only bias the weight of the first mode
        # The total initial "weight mass" in each row is reduced by 1792*0.02*0.5.
        self.key.weight.data[: self.attention_mode_dim] = (
            self.key.weight.data[: self.attention_mode_dim] * 0.5 + identity_weight
        )

    def transpose_for_scores(self, x):
        new_x_shape = x.size()[:-1] + (self.num_modes, -1)
        x = x.view(*new_x_shape)
        return x.permute(0, 2, 1, 3)

    # pos_biases: [1, 1, U1, U2].
    def forward(self, query_feat, key_feat=None, pos_biases=None, attention_mask=None):
        # query_feat: [B, U1, 1792]
        # if key_feat == None: self attention.
        if key_feat is None:
            key_feat = query_feat
        # mixed_query_layer, mixed_key_layer: [B, U1, 1792], [B, U2, 1792]
        mixed_query_layer = self.query(query_feat)
        mixed_key_layer = self.key(key_feat)
        # query_layer, key_layer: [B, 4, U1, 448], [B, 4, U2, 448]
        query_layer = self.transpose_for_scores(mixed_query_layer)
        key_layer = self.transpose_for_scores(mixed_key_layer)

        # Take the dot product between "query" and "key" to get the raw attention scores.
        attention_scores = torch.matmul(
            query_layer, key_layer.transpose(-1, -2)
        )  # [B0, 4, U1, 448] [B0, 4, 448, U2]
        attention_scores = attention_scores / math.sqrt(
            self.attention_mode_dim
        )  # [B0, 4, U1, U2]

        # if self.call_count == 0:
        #    print0(f"{self.name} query: {list(query_feat.shape)}, attn: {list(attention_scores.shape)}")

        with torch.no_grad():
            curr_max_attn = attention_scores.max().item()
            curr_avg_attn = attention_scores.abs().mean().item()

        if curr_max_attn > self.max_attn:
            self.max_attn = curr_max_attn

        if curr_max_attn > self.attn_clip:
            attention_scores = torch.clamp(
                attention_scores, -self.attn_clip, self.attn_clip
            )
            self.clamp_count += 1

        self.call_count += 1
        if self.training:
            if self.call_count % self.attn_diag_cycles == 0:
                print0(
                    "max-attn: {:.2f}, avg-attn: {:.2f}, clamp-count: {}".format(
                        self.max_attn, curr_avg_attn, self.clamp_count
                    )
                )
                self.max_attn = 0
                self.clamp_count = 0

        if pos_biases is not None:
            # [B0, 8, U1, U2] = [B0, 8, U1, U2]  + [1, 1, U1, U2].
            attention_scores = attention_scores + self.pos_code_weight * pos_biases
        if attention_mask is not None:
            attention_scores = attention_scores + attention_mask

        # When out_attn_scores_only, dropout is not applied to attention scores.
        if self.out_attn_scores_only:
            if self.num_modes > 1:
                # [3, num_modes=4, 4500, 4500] => [3, 1, 4500, 4500]
                attention_scores = self.attn_softaggr(attention_scores)
            # attention_scores = self.att_dropout(attention_scores)
            return attention_scores

        # Normalize the attention scores to probabilities.
        attention_probs = nn.functional.softmax(attention_scores, dim=-1)
        # self.attention_probs = attention_probs

        # lucidrains doesn't have this dropout but rwightman has. Will keep it.
        attention_probs = self.att_dropout(attention_probs)  # [B0, 4, U1, U2]

        if self.out_attn_probs_only:
            # [6, 4, 4500, 4500]
            return attention_probs

        else:
            # out_feat: [B0, U1, 1792], in the same size as query_feat.
            out_feat = self.out_trans(key_feat, attention_probs)
            return out_feat


class SelfAttVisPosTrans(nn.Module):
    def __init__(self, config, name):
        nn.Module.__init__(self)
        self.config = copy.copy(config)
        self.name = name
        self.out_attn_only = config.out_attn_scores_only or config.out_attn_probs_only
        self.attn_mask_radius = config.attn_mask_radius
        self.setrans = CrossAttFeatTrans(self.config, name)
        self.vispos_encoder = SETransInputFeatEncoder(self.config)

    def forward(self, x):
        coords = gen_all_indices(x.shape[2:], device=x.device)
        if self.attn_mask_radius > 0:
            coords2 = coords.reshape(-1, 2)
            coords_diff = coords2.unsqueeze(0) - coords2.unsqueeze(1)
            attn_mask = (coords_diff.abs().max(dim=2)[0] > self.attn_mask_radius).to(
                dtype=x.dtype
            )
            attn_mask = (attn_mask * -1e9).unsqueeze(0).unsqueeze(0)
        else:
            attn_mask = None

        coords = coords.unsqueeze(0).repeat(x.shape[0], 1, 1, 1)

        x_vispos, pos_biases = self.vispos_encoder(x, coords, return_pos_biases=True)

        # if out_attn_scores_only/out_attn_probs_only,
        # then x_trans is an attention matrix in the shape of (query unit number, key unit number)
        # otherwise, output features are in the same shape as the query features.
        # key features are recombined to get new query features by matmul(attention_probs, V(key features))
        #             frame1 frame2
        # x_vispos, x_trans: [4, 2852, 256]
        # Here key_feat is omitted (None), i.e., key_feat = query_feat = x_vispos.
        x_trans = self.setrans(
            x_vispos, pos_biases=pos_biases, attention_mask=attn_mask
        )

        # Save f2 attention for visualization
        if self.name == "F2 transformer" and "SAVEF2" in os.environ:
            # save the attention scores
            f2_attention_probs = self.setrans.attention_probs.detach().cpu()
            # [B0, 4, U1, U2] => [B0, U1, U2]
            f2_attention_probs = f2_attention_probs.mean(dim=1, keepdim=False)
            f2_savepath = os.environ["SAVEF2"]
            batch, C, h1, w1 = x.shape
            f2attn = f2_attention_probs.reshape(batch, h1, w1, h1, w1)
            torch.save(f2attn, f2_savepath)
            print0(f"F2 attention tensor saved to {f2_savepath}")

        # reshape x_trans to the input shape.
        if not self.out_attn_only:
            x_trans_shape = x_trans.shape
            x_trans = x_trans.permute(0, 2, 1).reshape(x.shape)

        return x_trans


# =================================== SETrans BackBone Components ==============================#


class LearnedSinuPosEmbedder(nn.Module):
    def __init__(self, pos_dim, pos_embed_dim, omega=1, affine=True):
        super().__init__()
        self.pos_dim = pos_dim
        self.pos_embed_dim = pos_embed_dim
        self.pos_fc = nn.Linear(self.pos_dim, self.pos_embed_dim, bias=True)
        self.pos_mix_norm_layer = nn.LayerNorm(
            self.pos_embed_dim, eps=1e-12, elementwise_affine=affine
        )
        self.omega = omega
        print0("Learnable Sinusoidal positional encoding")

    def forward(self, pos_normed):
        pos_embed_sum = 0
        pos_embed0 = self.pos_fc(pos_normed)
        pos_embed_sin = torch.sin(self.omega * pos_embed0[:, :, 0::2])
        pos_embed_cos = torch.cos(self.omega * pos_embed0[:, :, 1::2])
        # Interlace pos_embed_sin and pos_embed_cos.
        pos_embed_mix = torch.stack((pos_embed_sin, pos_embed_cos), dim=3).view(
            pos_embed0.shape
        )
        pos_embed_out = self.pos_mix_norm_layer(pos_embed_mix)

        return pos_embed_out


class SlidingPosBiases2D(nn.Module):
    def __init__(self, pos_dim=2, pos_bias_radius=7, max_pos_size=(200, 200)):
        super().__init__()
        self.pos_dim = pos_dim
        self.R = R = pos_bias_radius
        # biases: [15, 15]
        pos_bias_shape = [pos_bias_radius * 2 + 1 for i in range(pos_dim)]
        self.biases = Parameter(torch.zeros(pos_bias_shape))
        # Currently only feature maps with a 2D spatial shape (i.e., 2D images) are supported.
        if self.pos_dim == 2:
            all_h1s, all_w1s, all_h2s, all_w2s = [], [], [], []
            for i in range(max_pos_size[0]):
                i_h1s, i_w1s, i_h2s, i_w2s = [], [], [], []
                for j in range(max_pos_size[1]):
                    h1s, w1s, h2s, w2s = torch.meshgrid(
                        torch.tensor(i),
                        torch.tensor(j),
                        torch.arange(i, i + 2 * R + 1),
                        torch.arange(j, j + 2 * R + 1),
                        indexing="ij",
                    )
                    i_h1s.append(h1s)
                    i_w1s.append(w1s)
                    i_h2s.append(h2s)
                    i_w2s.append(w2s)

                i_h1s = torch.cat(i_h1s, dim=1)
                i_w1s = torch.cat(i_w1s, dim=1)
                i_h2s = torch.cat(i_h2s, dim=1)
                i_w2s = torch.cat(i_w2s, dim=1)
                all_h1s.append(i_h1s)
                all_w1s.append(i_w1s)
                all_h2s.append(i_h2s)
                all_w2s.append(i_w2s)

            all_h1s = torch.cat(all_h1s, dim=0)
            all_w1s = torch.cat(all_w1s, dim=0)
            all_h2s = torch.cat(all_h2s, dim=0)
            all_w2s = torch.cat(all_w2s, dim=0)
        else:
            breakpoint()

        # Put indices on GPU to speed up.
        # But if without persistent=False, they will be saved to checkpoints,
        # making the checkpoints unnecessarily huge.
        self.register_buffer("all_h1s", all_h1s, persistent=False)
        self.register_buffer("all_w1s", all_w1s, persistent=False)
        self.register_buffer("all_h2s", all_h2s, persistent=False)
        self.register_buffer("all_w2s", all_w2s, persistent=False)
        print0(f"Sliding-window Positional Biases, r: {R}, max size: {max_pos_size}")

    def forward(self, feat_shape, dtype, device):
        R = self.R
        spatial_shape = feat_shape[-self.pos_dim :]
        # [H, W, H, W] => [H+2R, W+2R, H+2R, W+2R].
        padded_pos_shape = list(spatial_shape) + [
            2 * R + spatial_shape[i] for i in range(self.pos_dim)
        ]
        padded_pos_biases = torch.zeros(padded_pos_shape, dtype=dtype, device=device)

        if self.pos_dim == 2:
            H, W = spatial_shape
            all_h1s = self.all_h1s[:H, :W]
            all_w1s = self.all_w1s[:H, :W]
            all_h2s = self.all_h2s[:H, :W]
            all_w2s = self.all_w2s[:H, :W]
            padded_pos_biases[(all_h1s, all_w1s, all_h2s, all_w2s)] = self.biases

        # Remove padding. [H+2R, W+2R, H+2R, W+2R] => [H, W, H, W].
        pos_biases = padded_pos_biases[:, :, R:-R, R:-R]

        return pos_biases


class SETransInputFeatEncoder(nn.Module):
    def __init__(self, config):
        super().__init__()
        self.feat_dim = config.in_feat_dim  # 256
        self.pos_embed_dim = self.feat_dim
        self.dropout = nn.Dropout(config.hidden_dropout_prob)
        self.comb_norm_layer = nn.LayerNorm(
            self.feat_dim, eps=1e-12, elementwise_affine=False
        )
        self.pos_code_type = config.pos_code_type

        # if using SlidingPosBiases2D, do not add positional embeddings here.
        if config.pos_code_type != "bias":
            self.pos_code_weight = config.pos_code_weight
            print0("Positional embedding weight: {:.3}".format(self.pos_code_weight))
        else:
            self.pos_code_weight = 0

        # Box position encoding. no affine, but could have bias.
        # 2 channels => 1792 channels
        if config.pos_code_type == "lsinu":
            self.pos_coder = LearnedSinuPosEmbedder(
                config.pos_dim, self.pos_embed_dim, omega=1, affine=False
            )
        elif config.pos_code_type == "rand":
            self.pos_coder = RandPosEmbedder(
                config.pos_dim, self.pos_embed_dim, shape=(36, 36), affine=False
            )
        elif config.pos_code_type == "sinu":
            self.pos_coder = SinuPosEmbedder(
                config.pos_dim, self.pos_embed_dim, shape=(36, 36), affine=False
            )
        elif config.pos_code_type == "zero":
            self.pos_coder = ZeroEmbedder(self.pos_embed_dim)
        elif config.pos_code_type == "bias":
            self.pos_coder = SlidingPosBiases2D(config.pos_dim, config.pos_bias_radius)

        self.cached_pos_code = None
        self.cached_feat_shape = None

    # Cache the pos_code and feat_shape to avoid unnecessary generation time.
    # This is only used during inference. During training, pos_code is always generated each time it's used.
    # Otherwise the cached pos_code cannot receive proper gradients.
    def pos_code_lookup_cache(self, vis_feat_shape, dtype, device, voxels_pos_normed):
        if self.pos_code_type == "bias":
            # Cache miss for 'bias' type of positional codes.
            if (
                self.training
                or self.cached_pos_code is None
                or self.cached_feat_shape != vis_feat_shape
            ):
                self.cached_pos_code = self.pos_coder(vis_feat_shape, dtype, device)
                self.cached_feat_shape = vis_feat_shape  # else: self.cached_pos_code exists, and self.cached_feat_shape == vis_feat_shape.
            # Just return the cached pos_code.
        else:
            # Cache miss for all other type of positional codes.
            if (
                self.training
                or self.cached_pos_code is None
                or self.cached_feat_shape != voxels_pos_normed.shape
            ):
                self.cached_pos_code = self.pos_coder(voxels_pos_normed)
                self.cached_feat_shape = voxels_pos_normed.shape
            # else: self.cached_pos_code exists, and self.cached_feat_shape == voxels_pos_normed.shape.
            # Just return the cached pos_code.
        return self.cached_pos_code

    # return: [B0, num_voxels, 256]
    def forward(self, vis_feat, voxels_pos, return_pos_biases=True):
        # vis_feat:  [8, 256, 46, 62]
        batch, dim, ht, wd = vis_feat.shape

        if self.pos_code_type != "bias":
            # voxels_pos: [8, 46, 62, 2]
            voxels_pos_normed = voxels_pos / voxels_pos.max()
            # voxels_pos_normed: [B0, num_voxels, 2]
            # pos_embed:         [B0, num_voxels, 256]
            voxels_pos_normed = voxels_pos_normed.view(batch, ht * wd, -1)
            pos_embed = self.pos_code_lookup_cache(
                vis_feat.shape, vis_feat.dtype, vis_feat.device, voxels_pos_normed
            )
            pos_biases = None
        else:
            pos_embed = 0
            # SlidingPosBiases2D() may be a bit slow. So only generate when necessary.
            if return_pos_biases:
                # pos_biases: [1, 1, H, W, H, W]
                pos_biases = self.pos_code_lookup_cache(
                    vis_feat.shape, vis_feat.dtype, vis_feat.device, None
                )
                # pos_biases: [1, 1, H*W, H*W]
                pos_biases = pos_biases.reshape(1, 1, ht * wd, ht * wd)
            else:
                # Simply discard pos_biases. Used when encoding the 2nd frame.
                # As for cross-frame attention, only one group of pos_biases is required
                # (added to the cross-frame attentio scores).
                # When encoding the 1st frame, pos_biases is already returned, no need
                # another group of pos_biases.
                pass

        vis_feat = vis_feat.view(batch, dim, ht * wd).transpose(1, 2)

        feat_comb = vis_feat + self.pos_code_weight * pos_embed
        feat_normed = self.comb_norm_layer(feat_comb)
        feat_normed = self.dropout(feat_normed)

        if return_pos_biases:
            return feat_normed, pos_biases
        else:
            return feat_normed



from typing import Sequence

import torch
import torch.nn as nn
import torch.nn.functional as F

from ptlflow.utils.registry import register_model, trainable
from .backbone import CNNEncoder
from .transformer import FeatureTransformer, FeatureFlowAttention
from .matching import global_correlation_softmax, local_correlation_softmax
from .geometry import flow_warp
from .utils import feature_add_position
from ..base_model.base_model import BaseModel


class SequenceLoss(nn.Module):
    def __init__(self, gamma: float, max_flow: float):
        super().__init__()
        self.gamma = gamma
        self.max_flow = max_flow

    def forward(self, outputs, inputs):
        """Loss function defined over sequence of flow predictions"""

        flow_preds = outputs["flow_preds"]
        flow_gt = inputs["flows"][:, 0]
        valid = inputs["valids"][:, 0]

        n_predictions = len(flow_preds)
        flow_loss = 0.0

        # exlude invalid pixels and extremely large diplacements
        mag = torch.sum(flow_gt**2, dim=1, keepdim=True).sqrt()
        valid = (valid >= 0.5) & (mag < self.max_flow)

        for i in range(n_predictions):
            i_weight = self.gamma ** (n_predictions - i - 1)

            i_loss = (flow_preds[i] - flow_gt).abs()

            flow_loss += i_weight * (valid[:, None] * i_loss).mean()

        return flow_loss


class GMFlow(BaseModel):
    pretrained_checkpoints = {
        "chairs": "https://github.com/hmorimitsu/ptlflow/releases/download/weights1/gmflow-chairs-4922131e.ckpt",
        "things": "https://github.com/hmorimitsu/ptlflow/releases/download/weights1/gmflow-things-5a18a9e8.ckpt",
        "sintel": "https://github.com/hmorimitsu/ptlflow/releases/download/weights1/gmflow-sintel-d6f83ccd.ckpt",
        "kitti": "https://github.com/hmorimitsu/ptlflow/releases/download/weights1/gmflow-kitti-af50eb2e.ckpt",
    }

    def __init__(
        self,
        attention_type: str = "swin",
        attn_splits_list: Sequence[int] = (2,),
        corr_radius_list: Sequence[int] = (-1,),
        feature_channels: int = 128,
        ffn_dim_expansion: int = 4,
        gamma: float = 0.9,
        max_flow: float = 400.0,
        num_head: int = 1,
        num_scales: int = 1,
        num_transformer_layers: int = 6,
        pred_bidir_flow: bool = False,
        prop_radius_list: Sequence[int] = (-1,),
        upsample_factor: int = 8,
        **kwargs,
    ) -> None:
        super().__init__(
            output_stride=32, loss_fn=SequenceLoss(gamma, max_flow), **kwargs
        )

        self.attention_type = attention_type
        self.attn_splits_list = attn_splits_list
        self.corr_radius_list = corr_radius_list
        self.feature_channels = feature_channels
        self.ffn_dim_expansion = ffn_dim_expansion
        self.gamma = gamma
        self.max_flow = max_flow
        self.num_head = num_head
        self.num_scales = num_scales
        self.num_transformer_layers = num_transformer_layers
        self.pred_bidir_flow = pred_bidir_flow
        self.prop_radius_list = prop_radius_list
        self.upsample_factor = upsample_factor

        # CNN backbone
        self.backbone = CNNEncoder(
            output_dim=self.feature_channels,
            num_output_scales=self.num_scales,
        )

        # Transformer
        self.transformer = FeatureTransformer(
            num_layers=self.num_transformer_layers,
            d_model=self.feature_channels,
            nhead=self.num_head,
            attention_type=self.attention_type,
            ffn_dim_expansion=self.ffn_dim_expansion,
        )

        # flow propagation with self-attn
        self.feature_flow_attn = FeatureFlowAttention(in_channels=self.feature_channels)

        # convex upsampling: concat feature0 and flow as input
        self.upsampler = nn.Sequential(
            nn.Conv2d(2 + self.feature_channels, 256, 3, 1, 1),
            nn.ReLU(inplace=True),
            nn.Conv2d(256, self.upsample_factor**2 * 9, 1, 1, 0),
        )

    def extract_feature(self, img0, img1):
        concat = torch.cat((img0, img1), dim=0)  # [2B, C, H, W]
        features = self.backbone(
            concat
        )  # list of [2B, C, H, W], resolution from high to low

        # reverse: resolution from low to high
        features = features[::-1]

        feature0, feature1 = [], []

        for i in range(len(features)):
            feature = features[i]
            chunks = torch.chunk(feature, 2, 0)  # tuple
            feature0.append(chunks[0])
            feature1.append(chunks[1])

        return feature0, feature1

    def upsample_flow(
        self,
        flow,
        feature,
        bilinear=False,
        upsample_factor=8,
    ):
        if bilinear:
            up_flow = (
                F.interpolate(
                    flow,
                    scale_factor=upsample_factor,
                    mode="bilinear",
                    align_corners=True,
                )
                * upsample_factor
            )

        else:
            # convex upsampling
            concat = torch.cat((flow, feature), dim=1)

            mask = self.upsampler(concat)
            b, flow_channel, h, w = flow.shape
            mask = mask.view(
                b, 1, 9, self.upsample_factor, self.upsample_factor, h, w
            )  # [B, 1, 9, K, K, H, W]
            mask = torch.softmax(mask, dim=2)

            up_flow = F.unfold(self.upsample_factor * flow, [3, 3], padding=1)
            up_flow = up_flow.view(
                b, flow_channel, 9, 1, 1, h, w
            )  # [B, 2, 9, 1, 1, H, W]

            up_flow = torch.sum(mask * up_flow, dim=2)  # [B, 2, K, K, H, W]
            up_flow = up_flow.permute(0, 1, 4, 2, 5, 3)  # [B, 2, K, H, K, W]
            up_flow = up_flow.reshape(
                b,
                flow_channel,
                self.upsample_factor * h,
                self.upsample_factor * w,
            )  # [B, 2, K*H, K*W]

        return up_flow

    def forward(self, inputs):
        """Estimate optical flow between pair of frames"""
        images, image_resizer = self.preprocess_images(
            inputs["images"],
            bgr_add=[-0.406, -0.456, -0.485],
            bgr_mult=[1 / 0.225, 1 / 0.224, 1 / 0.229],
            bgr_to_rgb=True,
            resize_mode="pad",
            pad_mode="replicate",
            pad_two_side=True,
        )

        img0 = images[:, 0]
        img1 = images[:, 1]

        flow_preds = []

        # resolution low to high
        feature0_list, feature1_list = self.extract_feature(
            img0, img1
        )  # list of features

        flow = None

        assert (
            len(self.attn_splits_list)
            == len(self.corr_radius_list)
            == len(self.prop_radius_list)
            == self.num_scales
        )

        for scale_idx in range(self.num_scales):
            feature0, feature1 = feature0_list[scale_idx], feature1_list[scale_idx]

            if self.pred_bidir_flow and scale_idx > 0:
                # predicting bidirectional flow with refinement
                feature0, feature1 = torch.cat((feature0, feature1), dim=0), torch.cat(
                    (feature1, feature0), dim=0
                )

            upsample_factor = self.upsample_factor * (
                2 ** (self.num_scales - 1 - scale_idx)
            )

            if scale_idx > 0:
                flow = (
                    F.interpolate(
                        flow, scale_factor=2, mode="bilinear", align_corners=True
                    )
                    * 2
                )

            if flow is not None:
                flow = flow.detach()
                feature1 = flow_warp(feature1, flow)  # [B, C, H, W]

            attn_splits = self.attn_splits_list[scale_idx]
            corr_radius = self.corr_radius_list[scale_idx]
            prop_radius = self.prop_radius_list[scale_idx]

            # add position to features
            feature0, feature1 = feature_add_position(
                feature0, feature1, attn_splits, self.feature_channels
            )

            # Transformer
            feature0, feature1 = self.transformer(
                feature0, feature1, attn_num_splits=attn_splits
            )

            # correlation and softmax
            if corr_radius == -1:  # global matching
                flow_pred = global_correlation_softmax(
                    feature0, feature1, self.pred_bidir_flow
                )[0]
            else:  # local matching
                flow_pred = local_correlation_softmax(feature0, feature1, corr_radius)[
                    0
                ]

            # flow or residual flow
            flow = flow + flow_pred if flow is not None else flow_pred

            # upsample to the original resolution for supervison
            if (
                self.training
            ):  # only need to upsample intermediate flow predictions at training time
                flow_bilinear = self.upsample_flow(
                    flow, None, bilinear=True, upsample_factor=upsample_factor
                )
                flow_bilinear = self.postprocess_predictions(
                    flow_bilinear, image_resizer, is_flow=True
                )
                flow_preds.append(flow_bilinear)

            # flow propagation with self-attn
            if self.pred_bidir_flow and scale_idx == 0:
                feature0 = torch.cat(
                    (feature0, feature1), dim=0
                )  # [2*B, C, H, W] for propagation
            flow = self.feature_flow_attn(
                feature0,
                flow.detach(),
                local_window_attn=prop_radius > 0,
                local_window_radius=prop_radius,
            )

            # bilinear upsampling at training time except the last one
            if self.training and scale_idx < self.num_scales - 1:
                flow_up = self.upsample_flow(
                    flow, feature0, bilinear=True, upsample_factor=upsample_factor
                )
                flow_up = self.postprocess_predictions(
                    flow_up, image_resizer, is_flow=True
                )
                flow_preds.append(flow_up)

            if scale_idx == self.num_scales - 1:
                flow_up = self.upsample_flow(flow, feature0)
                flow_up = self.postprocess_predictions(
                    flow_up, image_resizer, is_flow=True
                )
                flow_preds.append(flow_up)

        if self.training:
            outputs = {"flows": flow_up[:, None], "flow_preds": flow_preds}
        else:
            outputs = {"flows": flow_up[:, None]}

        return outputs


class GMFlowWithRefinement(GMFlow):
    pretrained_checkpoints = {
        "chairs": "https://github.com/hmorimitsu/ptlflow/releases/download/weights1/gmflow_refine-chairs-88cdc009.ckpt",
        "things": "https://github.com/hmorimitsu/ptlflow/releases/download/weights1/gmflow_refine-things-e40899f5.ckpt",
        "sintel": "https://github.com/hmorimitsu/ptlflow/releases/download/weights1/gmflow_refine-sintel-ee46a2c4.ckpt",
        "kitti": "https://github.com/hmorimitsu/ptlflow/releases/download/weights1/gmflow_refine-kitti-b7bf2fda.ckpt",
    }

    def __init__(
        self,
        attention_type: str = "swin",
        attn_splits_list: Sequence[int] = (2, 8),
        corr_radius_list: Sequence[int] = (-1, 4),
        feature_channels: int = 128,
        ffn_dim_expansion: int = 4,
        gamma: float = 0.9,
        max_flow: float = 400,
        num_head: int = 1,
        num_scales: int = 2,
        num_transformer_layers: int = 6,
        pred_bidir_flow: bool = False,
        prop_radius_list: Sequence[int] = (-1, 1),
        upsample_factor: int = 4,
        **kwargs,
    ) -> None:
        super().__init__(
            attention_type,
            attn_splits_list,
            corr_radius_list,
            feature_channels,
            ffn_dim_expansion,
            gamma,
            max_flow,
            num_head,
            num_scales,
            num_transformer_layers,
            pred_bidir_flow,
            prop_radius_list,
            upsample_factor,
            **kwargs,
        )


@register_model
@trainable
class gmflow(GMFlow):
    pass


@register_model
@trainable
class gmflow_refine(GMFlowWithRefinement):
    pass



import torch
import torch.nn.functional as F

# from spatial_correlation_sampler import SpatialCorrelationSampler


class UpSample(torch.nn.Module):
    def __init__(self, feature_dim, upsample_factor):
        super(UpSample, self).__init__()

        self.upsample_factor = upsample_factor

        self.conv1 = torch.nn.Conv2d(2 + feature_dim, 256, 3, 1, 1)
        self.conv2 = torch.nn.Conv2d(256, 512, 3, 1, 1)
        self.conv3 = torch.nn.Conv2d(512, upsample_factor**2 * 9, 1, 1, 0)
        self.relu = torch.nn.ReLU(inplace=True)

    def forward(self, feature, flow):
        concat = torch.cat((flow, feature), dim=1)

        mask = self.conv3(self.relu(self.conv2(self.relu(self.conv1(concat)))))

        b, _, h, w = flow.shape

        mask = mask.view(
            b, 1, 9, self.upsample_factor, self.upsample_factor, h, w
        )  # [B, 1, 9, K, K, H, W]
        mask = torch.softmax(mask, dim=2)

        # up_flow = F.unfold(self.upsample_factor * flow, [3, 3], padding=1)
        up_flow = F.unfold(flow, [3, 3], padding=1)
        up_flow = up_flow.view(b, 2, 9, 1, 1, h, w)  # [B, 2, 9, 1, 1, H, W]

        up_flow = torch.sum(mask * up_flow, dim=2)  # [B, 2, K, K, H, W]
        up_flow = up_flow.permute(0, 1, 4, 2, 5, 3)  # [B, 2, K, H, K, W]
        up_flow = up_flow.reshape(
            b, 2, self.upsample_factor * h, self.upsample_factor * w
        )  # [B, 2, K*H, K*W]

        return up_flow



"""Validate optical flow estimation performance on standard datasets."""

# =============================================================================
# Copyright 2024 Henrique Morimitsu
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# =============================================================================

import sys
from pathlib import Path

from jsonargparse import ArgumentParser
from loguru import logger
import torch
import torch.onnx

this_dir = Path(__file__).parent.resolve()
sys.path.insert(0, str(this_dir.parent.parent.parent))

from ptlflow import get_model, load_checkpoint
from ptlflow.utils.lightning.ptlflow_cli import PTLFlowCLI
from ptlflow.utils.registry import RegisteredModel


def _init_parser() -> ArgumentParser:
    parser = ArgumentParser(add_help=False)
    parser.add_argument(
        "--ckpt_path",
        type=str,
        default=None,
        help="Path to the checkpoint to be loaded. It can also be one of the following names: {chairs, things, sintel, kitti}, in which case the respective pretrained checkpoint will be downloaded.",
    )
    parser.add_argument(
        "--output_path",
        type=str,
        default=".",
        help="Path to the directory where the converted onnx model will be saved.",
    )
    parser.add_argument(
        "--input_size",
        type=int,
        nargs=2,
        default=[384, 1280],
        help="Size of the input image.",
    )
    return parser


def fuse_checkpoint_next1d_layers(state_dict):
    fused_sd = {}
    hv_pairs = {}
    for name, param in state_dict.items():
        if name.endswith("weight_h") or name.endswith("weight_v"):
            name_prefix = name[: -(len("weight_h") + 1)]
            orientation = name[-1]
            if name_prefix not in hv_pairs:
                hv_pairs[name_prefix] = {}
            hv_pairs[name_prefix][orientation] = param
        else:
            fused_sd[name] = param

    for name_prefix, param_pairs in hv_pairs.items():
        weight = torch.einsum("cijk,cimj->cimk", param_pairs["h"], param_pairs["v"])
        fused_sd[f"{name_prefix}.weight"] = weight
    return fused_sd


def load_model(cfg):
    ckpt_path = cfg.ckpt_path
    cfg.ckpt_path = None  # This is to avoid letting get_model to load the ckpt
    model = get_model(cfg.model_name, args=cfg)
    # Since we set fuse_next1d_weights to True in the model,
    # the ckpt 1D layers also need to be fused before loading
    ckpt = load_checkpoint(ckpt_path, model.__class__)
    state_dict = fuse_checkpoint_next1d_layers(ckpt["state_dict"])
    model.load_state_dict(state_dict, strict=True)
    return model


def _show_v04_warning():
    ignore_args = ["-h", "--help", "--model", "--config"]
    for arg in ignore_args:
        if arg in sys.argv:
            return

    logger.warning(
        "Since v0.4, it is now necessary to inform the model using the --model argument. For example, use: python infer.py --model raft --ckpt_path things"
    )


if __name__ == "__main__":
    _show_v04_warning()

    parser = _init_parser()

    cli = PTLFlowCLI(
        model_class=RegisteredModel,
        subclass_mode_model=True,
        parser_kwargs={"parents": [parser]},
        run=False,
        parse_only=False,
        auto_configure_optimizers=False,
    )

    cfg = cli.config
    cfg.model.init_args.corr_mode = "allpairs"
    cfg.model.init_args.fuse_next1d_weights = True
    cfg.model.init_args.simple_io = True
    cfg.model_name = cfg.model.class_path.split(".")[-1]

    model = load_model(cfg)

    # model = get_model(cfg.model_name, args=cfg)
    sample_inputs = torch.randn(1, 2, 3, cfg.input_size[0], cfg.input_size[1])
    if torch.cuda.is_available():
        model = model.cuda()
        sample_inputs = sample_inputs.cuda()

    output_dir = Path(cfg.output_path)
    output_dir.mkdir(parents=True, exist_ok=True)
    output_path = str(output_dir / f"{cfg.model_name}.onnx")
    torch.onnx.export(
        model, sample_inputs, output_path, verbose=False, opset_version=16
    )
    print(f"ONNX model saved to: {output_path}")



import torch
import torch.nn.functional as F


class TransformerLayer(torch.nn.Module):
    def __init__(self, feature_dim, ffn=True, ffn_dim_expansion=1):
        super(TransformerLayer, self).__init__()

        # multi-head attention
        self.q_proj = torch.nn.Linear(feature_dim, feature_dim)
        self.k_proj = torch.nn.Linear(feature_dim, feature_dim)
        self.v_proj = torch.nn.Linear(feature_dim, feature_dim)

        self.merge = torch.nn.Linear(feature_dim, feature_dim)

        # self.multi_head_attn = torch.nn.MultiheadAttention(feature_dim, 2, batch_first=True, device='cuda')

        self.norm1 = torch.nn.LayerNorm(feature_dim)

        self.ffn = ffn

        if self.ffn:
            in_channels = feature_dim * 2
            self.mlp = torch.nn.Sequential(
                torch.nn.Linear(
                    in_channels, in_channels * ffn_dim_expansion, bias=False
                ),
                torch.nn.GELU(),
                torch.nn.Linear(
                    in_channels * ffn_dim_expansion, feature_dim, bias=False
                ),
            )

            self.norm2 = torch.nn.LayerNorm(feature_dim)

    def forward(self, source, target):
        # source, target: [B, L, C]
        query, key, value = source, target, target

        # single-head attention
        query = self.q_proj(query)  # [B, L, C]
        key = self.k_proj(key)  # [B, L, C]
        value = self.v_proj(value)  # [B, L, C]

        message = F.scaled_dot_product_attention(query, key, value, dropout_p=0.0)

        message = self.merge(message)

        # message, _ = self.multi_head_attn(query, key, value, need_weights=False)
        message = self.norm1(message)

        if self.ffn:
            message = self.mlp(torch.cat([source, message], dim=-1))
            message = self.norm2(message)

        return source + message


class FeatureAttention(torch.nn.Module):
    def __init__(
        self,
        feature_dim,
        num_layers,
        bidir=True,
        ffn=True,
        ffn_dim_expansion=1,
        post_norm=False,
    ):
        super(FeatureAttention, self).__init__()

        self.bidir = bidir

        self.layers = torch.nn.ModuleList(
            [
                TransformerLayer(
                    feature_dim, ffn=ffn, ffn_dim_expansion=ffn_dim_expansion
                )
                for i in range(num_layers)
            ]
        )

        self.post_norm = post_norm

        if self.post_norm:
            self.norm = torch.nn.LayerNorm(feature_dim, eps=1e-06)

    def forward(self, feature0, feature1):
        b, c, h, w = feature0.shape

        feature0 = feature0.flatten(-2).permute(0, 2, 1)  # [B, H*W, C]
        feature1 = feature1.flatten(-2).permute(0, 2, 1)  # [B, H*W, C]

        if self.bidir:
            concat0 = torch.cat((feature0, feature1), dim=0)  # [2B, H*W, C]
            concat1 = torch.cat((feature1, feature0), dim=0)  # [2B, H*W, C]

            for layer in self.layers:
                concat0 = layer(concat0, concat1)
                concat1 = torch.cat(concat0.chunk(chunks=2, dim=0)[::-1], dim=0)

            if self.post_norm:
                concat0 = self.norm(concat0)

            feature0, feature1 = concat0.chunk(chunks=2, dim=0)  # [B, H*W, C]

            # reshape back
            feature0 = (
                feature0.view(b, h, w, c).permute(0, 3, 1, 2).contiguous()
            )  # [B, C, H, W]
            feature1 = (
                feature1.view(b, h, w, c).permute(0, 3, 1, 2).contiguous()
            )  # [B, C, H, W]

            return feature0, feature1

        else:
            for layer in self.layers:
                feature0 = layer(feature0, feature1)

            if self.post_norm:
                feature0 = self.norm(feature0)

            feature0 = (
                feature0.view(b, h, w, c).permute(0, 3, 1, 2).contiguous()
            )  # [B, C, H, W]

            return feature0


class FlowAttention(torch.nn.Module):
    """
    flow propagation with self-attention on feature
    query: feature0, key: feature0, value: flow
    """

    def __init__(self, feature_dim):
        super(FlowAttention, self).__init__()

        self.q_proj = torch.nn.Linear(feature_dim, feature_dim)
        self.k_proj = torch.nn.Linear(feature_dim, feature_dim)

    def forward(self, feature, flow):
        # q, k: feature [B, C, H, W], v: flow [B, 2, H, W]
        b, c, h, w = feature.size()

        feature = feature.flatten(-2).permute(0, 2, 1)  # [B, H*W, C]

        flow = flow.flatten(-2).permute(0, 2, 1)

        query = self.q_proj(feature)  # [B, H*W, C]
        key = self.k_proj(feature)  # [B, H*W, C]

        flow = F.scaled_dot_product_attention(query, key, flow)

        flow = flow.view(b, h, w, 2).permute(0, 3, 1, 2)

        return flow



import torch
import torch.nn as nn
import torch.nn.functional as F

from .utils import split_feature, merge_splits, split_feature_1d, merge_splits_1d


def single_head_full_attention(q, k, v):
    # q, k, v: [B, L, C]
    assert q.dim() == k.dim() == v.dim() == 3

    scores = torch.matmul(q, k.permute(0, 2, 1)) / (q.size(2) ** 0.5)  # [B, L, L]
    attn = torch.softmax(scores, dim=2)  # [B, L, L]
    out = torch.matmul(attn, v)  # [B, L, C]

    return out


def single_head_full_attention_1d(
    q,
    k,
    v,
    h=None,
    w=None,
):
    # q, k, v: [B, L, C]

    assert h is not None and w is not None
    assert q.size(1) == h * w

    b, _, c = q.size()

    q = q.view(b, h, w, c)  # [B, H, W, C]
    k = k.view(b, h, w, c)
    v = v.view(b, h, w, c)

    scale_factor = c**0.5

    scores = torch.matmul(q, k.permute(0, 1, 3, 2)) / scale_factor  # [B, H, W, W]

    attn = torch.softmax(scores, dim=-1)

    out = torch.matmul(attn, v).view(b, -1, c)  # [B, H*W, C]

    return out


def single_head_split_window_attention(
    q,
    k,
    v,
    num_splits=1,
    with_shift=False,
    h=None,
    w=None,
    attn_mask=None,
):
    # ref: https://github.com/microsoft/Swin-Transformer/blob/main/models/swin_transformer.py
    # q, k, v: [B, L, C]
    assert q.dim() == k.dim() == v.dim() == 3

    assert h is not None and w is not None
    assert q.size(1) == h * w

    b, _, c = q.size()

    b_new = b * num_splits * num_splits

    window_size_h = h // num_splits
    window_size_w = w // num_splits

    q = q.view(b, h, w, c)  # [B, H, W, C]
    k = k.view(b, h, w, c)
    v = v.view(b, h, w, c)

    scale_factor = c**0.5

    if with_shift:
        assert attn_mask is not None  # compute once
        shift_size_h = window_size_h // 2
        shift_size_w = window_size_w // 2

        q = torch.roll(q, shifts=(-shift_size_h, -shift_size_w), dims=(1, 2))
        k = torch.roll(k, shifts=(-shift_size_h, -shift_size_w), dims=(1, 2))
        v = torch.roll(v, shifts=(-shift_size_h, -shift_size_w), dims=(1, 2))

    q = split_feature(
        q, num_splits=num_splits, channel_last=True
    )  # [B*K*K, H/K, W/K, C]
    k = split_feature(k, num_splits=num_splits, channel_last=True)
    v = split_feature(v, num_splits=num_splits, channel_last=True)

    scores = (
        torch.matmul(q.view(b_new, -1, c), k.view(b_new, -1, c).permute(0, 2, 1))
        / scale_factor
    )  # [B*K*K, H/K*W/K, H/K*W/K]

    if with_shift:
        scores += attn_mask.repeat(b, 1, 1)

    attn = torch.softmax(scores, dim=-1)

    out = torch.matmul(attn, v.view(b_new, -1, c))  # [B*K*K, H/K*W/K, C]

    out = merge_splits(
        out.view(b_new, h // num_splits, w // num_splits, c),
        num_splits=num_splits,
        channel_last=True,
    )  # [B, H, W, C]

    # shift back
    if with_shift:
        out = torch.roll(out, shifts=(shift_size_h, shift_size_w), dims=(1, 2))

    out = out.view(b, -1, c)

    return out


def single_head_split_window_attention_1d(
    q,
    k,
    v,
    relative_position_bias=None,
    num_splits=1,
    with_shift=False,
    h=None,
    w=None,
    attn_mask=None,
):
    # q, k, v: [B, L, C]

    assert h is not None and w is not None
    assert q.size(1) == h * w

    b, _, c = q.size()

    b_new = b * num_splits * h

    window_size_w = w // num_splits

    q = q.view(b * h, w, c)  # [B*H, W, C]
    k = k.view(b * h, w, c)
    v = v.view(b * h, w, c)

    scale_factor = c**0.5

    if with_shift:
        assert attn_mask is not None  # compute once
        shift_size_w = window_size_w // 2

        q = torch.roll(q, shifts=-shift_size_w, dims=1)
        k = torch.roll(k, shifts=-shift_size_w, dims=1)
        v = torch.roll(v, shifts=-shift_size_w, dims=1)

    q = split_feature_1d(q, num_splits=num_splits)  # [B*H*K, W/K, C]
    k = split_feature_1d(k, num_splits=num_splits)
    v = split_feature_1d(v, num_splits=num_splits)

    scores = (
        torch.matmul(q.view(b_new, -1, c), k.view(b_new, -1, c).permute(0, 2, 1))
        / scale_factor
    )  # [B*H*K, W/K, W/K]

    if with_shift:
        # attn_mask: [K, W/K, W/K]
        scores += attn_mask.repeat(b * h, 1, 1)  # [B*H*K, W/K, W/K]

    attn = torch.softmax(scores, dim=-1)

    out = torch.matmul(attn, v.view(b_new, -1, c))  # [B*H*K, W/K, C]

    out = merge_splits_1d(out, h, num_splits=num_splits)  # [B, H, W, C]

    # shift back
    if with_shift:
        out = torch.roll(out, shifts=shift_size_w, dims=2)

    out = out.view(b, -1, c)

    return out


class SelfAttnPropagation(nn.Module):
    """
    flow propagation with self-attention on feature
    query: feature0, key: feature0, value: flow
    """

    def __init__(
        self,
        in_channels,
        **kwargs,
    ):
        super(SelfAttnPropagation, self).__init__()

        self.q_proj = nn.Linear(in_channels, in_channels)
        self.k_proj = nn.Linear(in_channels, in_channels)

        for p in self.parameters():
            if p.dim() > 1:
                nn.init.xavier_uniform_(p)

    def forward(
        self,
        feature0,
        flow,
        local_window_attn=False,
        local_window_radius=1,
        **kwargs,
    ):
        # q, k: feature [B, C, H, W], v: flow [B, 2, H, W]
        if local_window_attn:
            return self.forward_local_window_attn(
                feature0, flow, local_window_radius=local_window_radius
            )

        b, c, h, w = feature0.size()

        query = feature0.view(b, c, h * w).permute(0, 2, 1)  # [B, H*W, C]

        # a note: the ``correct'' implementation should be:
        # ``query = self.q_proj(query), key = self.k_proj(query)''
        # this problem is observed while cleaning up the code
        # however, this doesn't affect the performance since the projection is a linear operation,
        # thus the two projection matrices for key can be merged
        # so I just leave it as is in order to not re-train all models :)
        query = self.q_proj(query)  # [B, H*W, C]
        key = self.k_proj(query)  # [B, H*W, C]

        value = flow.view(b, flow.size(1), h * w).permute(0, 2, 1)  # [B, H*W, 2]

        scores = torch.matmul(query, key.permute(0, 2, 1)) / (c**0.5)  # [B, H*W, H*W]
        prob = torch.softmax(scores, dim=-1)

        out = torch.matmul(prob, value)  # [B, H*W, 2]
        out = out.view(b, h, w, value.size(-1)).permute(0, 3, 1, 2)  # [B, 2, H, W]

        return out

    def forward_local_window_attn(
        self,
        feature0,
        flow,
        local_window_radius=1,
    ):
        assert flow.size(1) == 2 or flow.size(1) == 1  # flow or disparity or depth
        assert local_window_radius > 0

        b, c, h, w = feature0.size()

        value_channel = flow.size(1)

        feature0_reshape = self.q_proj(
            feature0.view(b, c, -1).permute(0, 2, 1)
        ).reshape(
            b * h * w, 1, c
        )  # [B*H*W, 1, C]

        kernel_size = 2 * local_window_radius + 1

        feature0_proj = (
            self.k_proj(feature0.view(b, c, -1).permute(0, 2, 1))
            .permute(0, 2, 1)
            .reshape(b, c, h, w)
        )

        feature0_window = F.unfold(
            feature0_proj, kernel_size=kernel_size, padding=local_window_radius
        )  # [B, C*(2R+1)^2), H*W]

        feature0_window = (
            feature0_window.view(b, c, kernel_size**2, h, w)
            .permute(0, 3, 4, 1, 2)
            .reshape(b * h * w, c, kernel_size**2)
        )  # [B*H*W, C, (2R+1)^2]

        flow_window = F.unfold(
            flow, kernel_size=kernel_size, padding=local_window_radius
        )  # [B, 2*(2R+1)^2), H*W]

        flow_window = (
            flow_window.view(b, value_channel, kernel_size**2, h, w)
            .permute(0, 3, 4, 2, 1)
            .reshape(b * h * w, kernel_size**2, value_channel)
        )  # [B*H*W, (2R+1)^2, 2]

        scores = torch.matmul(feature0_reshape, feature0_window) / (
            c**0.5
        )  # [B*H*W, 1, (2R+1)^2]

        prob = torch.softmax(scores, dim=-1)

        out = (
            torch.matmul(prob, flow_window)
            .view(b, h, w, value_channel)
            .permute(0, 3, 1, 2)
            .contiguous()
        )  # [B, 2, H, W]

        return out



# Copyright The Lightning AI team.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import sys
from typing import Any, Callable, Dict, Optional, Type, Union

from jsonargparse import Namespace
from lightning import LightningDataModule, LightningModule, Trainer
from lightning.pytorch.cli import (
    ArgsType,
    LightningArgumentParser,
    LightningCLI,
    SaveConfigCallback,
)
from lightning.pytorch.utilities.rank_zero import rank_zero_warn


class PTLFlowCLI(LightningCLI):
    def __init__(
        self,
        model_class: Optional[
            Union[Type[LightningModule], Callable[..., LightningModule]]
        ] = None,
        datamodule_class: Optional[
            Union[Type[LightningDataModule], Callable[..., LightningDataModule]]
        ] = None,
        save_config_callback: Optional[Type[SaveConfigCallback]] = SaveConfigCallback,
        save_config_kwargs: Optional[Dict[str, Any]] = None,
        trainer_class: Optional[Union[Type[Trainer], Callable[..., Trainer]]] = None,
        trainer_defaults: Optional[Dict[str, Any]] = None,
        seed_everything_default: Union[bool, int] = True,
        parser_kwargs: Optional[
            Union[Dict[str, Any], Dict[str, Dict[str, Any]]]
        ] = None,
        subclass_mode_model: bool = False,
        subclass_mode_data: bool = False,
        args: ArgsType = None,
        run: bool = True,
        auto_configure_optimizers: bool = True,
        parse_only: bool = False,
        ignore_sys_argv: bool = False,
    ) -> None:
        """Receives as input pytorch-lightning classes (or callables which return pytorch-lightning classes), which are
        called / instantiated using a parsed configuration file and / or command line args.

        Parsing of configuration from environment variables can be enabled by setting ``parser_kwargs={"default_env":
        True}``. A full configuration yaml would be parsed from ``PL_CONFIG`` if set. Individual settings are so parsed
        from variables named for example ``PL_TRAINER__MAX_EPOCHS``.

        For more info, read :ref:`the CLI docs <lightning-cli>`.

        Args:
            model_class: An optional :class:`~lightning.pytorch.core.LightningModule` class to train on or a
                callable which returns a :class:`~lightning.pytorch.core.LightningModule` instance when
                called. If ``None``, you can pass a registered model with ``--model=MyModel``.
            datamodule_class: An optional :class:`~lightning.pytorch.core.datamodule.LightningDataModule` class or a
                callable which returns a :class:`~lightning.pytorch.core.datamodule.LightningDataModule` instance when
                called. If ``None``, you can pass a registered datamodule with ``--data=MyDataModule``.
            save_config_callback: A callback class to save the config.
            save_config_kwargs: Parameters that will be used to instantiate the save_config_callback.
            trainer_class: An optional subclass of the :class:`~lightning.pytorch.trainer.trainer.Trainer` class or a
                callable which returns a :class:`~lightning.pytorch.trainer.trainer.Trainer` instance when called.
            trainer_defaults: Set to override Trainer defaults or add persistent callbacks. The callbacks added through
                this argument will not be configurable from a configuration file and will always be present for
                this particular CLI. Alternatively, configurable callbacks can be added as explained in
                :ref:`the CLI docs <lightning-cli>`.
            seed_everything_default: Number for the :func:`~lightning.fabric.utilities.seed.seed_everything`
                seed value. Set to True to automatically choose a seed value.
                Setting it to False will avoid calling ``seed_everything``.
            parser_kwargs: Additional arguments to instantiate each ``LightningArgumentParser``.
            subclass_mode_model: Whether model can be any `subclass
                <https://jsonargparse.readthedocs.io/en/stable/#class-type-and-sub-classes>`_
                of the given class.
            subclass_mode_data: Whether datamodule can be any `subclass
                <https://jsonargparse.readthedocs.io/en/stable/#class-type-and-sub-classes>`_
                of the given class.
            args: Arguments to parse. If ``None`` the arguments are taken from ``sys.argv``. Command line style
                arguments can be given in a ``list``. Alternatively, structured config options can be given in a
                ``dict`` or ``jsonargparse.Namespace``.
            run: Whether subcommands should be added to run a :class:`~lightning.pytorch.trainer.trainer.Trainer`
                method. If set to ``False``, the trainer and model classes will be instantiated only.
            parse_only: If set to ``True``, the CLI acts as a parser only. The classes are not instantiated.
        """
        self.save_config_callback = save_config_callback
        self.save_config_kwargs = save_config_kwargs or {}
        self.trainer_class = trainer_class
        self.trainer_defaults = trainer_defaults or {}
        self.seed_everything_default = seed_everything_default
        self.parser_kwargs = parser_kwargs or {}
        self.auto_configure_optimizers = auto_configure_optimizers

        self.model_class = model_class
        # used to differentiate between the original value and the processed value
        self._model_class = model_class
        self.subclass_mode_model = subclass_mode_model

        self.datamodule_class = datamodule_class
        # used to differentiate between the original value and the processed value
        self._datamodule_class = datamodule_class
        self.subclass_mode_data = (datamodule_class is None) or subclass_mode_data

        main_kwargs, subparser_kwargs = self._setup_parser_kwargs(self.parser_kwargs)
        self.setup_parser(run, main_kwargs, subparser_kwargs)
        self.parse_arguments(self.parser, args, ignore_sys_argv)

        if not parse_only:
            self.subcommand = self.config["subcommand"] if run else None

            self._set_seed()

            self._add_instantiators()
            self.before_instantiate_classes()
            self.instantiate_classes()

            if self.subcommand is not None:
                self._run_subcommand(self.subcommand)

    def add_core_arguments_to_parser(self, parser: LightningArgumentParser) -> None:
        """Adds arguments from the core classes to the parser."""
        if self.trainer_class is not None:
            parser.add_lightning_class_args(self.trainer_class, "trainer")
            trainer_defaults = {
                "trainer." + k: v
                for k, v in self.trainer_defaults.items()
                if k != "callbacks"
            }
            parser.set_defaults(trainer_defaults)

        if self._model_class is not None:
            parser.add_lightning_class_args(
                self._model_class, "model", subclass_mode=self.subclass_mode_model
            )

        if self.datamodule_class is not None:
            parser.add_lightning_class_args(
                self._datamodule_class, "data", subclass_mode=self.subclass_mode_data
            )

    def parse_arguments(
        self, parser: LightningArgumentParser, args: ArgsType, ignore_sys_argv: bool
    ) -> None:
        """Parses command line arguments and stores it in ``self.config``."""
        if args is not None and len(sys.argv) > 1 and not ignore_sys_argv:
            rank_zero_warn(
                "LightningCLI's args parameter is intended to run from within Python like if it were from the command "
                "line. To prevent mistakes it is not recommended to provide both args and command line arguments, got: "
                f"sys.argv[1:]={sys.argv[1:]}, args={args}."
            )
        if isinstance(args, (dict, Namespace)):
            self.config = parser.parse_object(args)
        else:
            self.config = parser.parse_args(args)

    def instantiate_classes(self) -> None:
        """Instantiates the classes and sets their attributes."""
        self.config_init = self.parser.instantiate_classes(self.config)
        self.datamodule = self._get(self.config_init, "data")
        self.model = self._get(self.config_init, "model")
        self._add_configure_optimizers_method_to_model(self.subcommand)

        if self.trainer_class is not None:
            self.trainer = self.instantiate_trainer()



from typing import Sequence

try:
    from spatial_correlation_sampler import SpatialCorrelationSampler
except ModuleNotFoundError:
    from ptlflow.utils.correlation import (
        IterSpatialCorrelationSampler as SpatialCorrelationSampler,
    )
import torch
import torch.nn as nn
import torch.nn.functional as F

from ptlflow.utils.registry import register_model, trainable
from .hd3_ops import *
from .dla_up import DLAEncoder as dlaup_encoder
from .vgg import VGGEncoder as vgg_encoder
from . import decoder as PreDecoder
from .hd3losses import LossCalculator
from ..base_model.base_model import BaseModel

BatchNorm = nn.BatchNorm2d


class Context(nn.Module):
    def __init__(self, inplane, classes):
        super(Context, self).__init__()
        self.num_convs = 7
        ch = [inplane, 128, 128, 128, 128, 128, 128, 128]
        dilations = [1, 1, 2, 4, 8, 16, 1]
        for i in range(self.num_convs):
            setattr(
                self,
                "dc_conv_{}".format(i),
                nn.Sequential(
                    nn.Conv2d(
                        ch[i],
                        ch[i + 1],
                        kernel_size=3,
                        stride=1,
                        padding=dilations[i],
                        dilation=dilations[i],
                        bias=False,
                    ),
                    BatchNorm(ch[i + 1]),
                    nn.ReLU(inplace=True),
                ),
            )
        self.cls = nn.Conv2d(
            ch[-1], classes, kernel_size=1, stride=1, padding=0, bias=True
        )

    def forward(self, x):
        out = x
        for i in range(self.num_convs):
            dc_conv = getattr(self, "dc_conv_" + str(i))
            out = dc_conv(out)
        out = self.cls(out)
        return out, None


class Decoder(nn.Module):
    def __init__(self, inplane, block, classes, up_classes):
        super(Decoder, self).__init__()
        self.mapping = block(inplane, 128)
        self.cls = nn.Sequential(
            BatchNorm(128),
            nn.ReLU(inplace=True),
            nn.Conv2d(128, classes, kernel_size=1, stride=1, padding=0, bias=True),
        )
        self.up = None
        if up_classes > 0:
            self.up = nn.Sequential(
                BatchNorm(128),
                nn.ReLU(inplace=True),
                nn.ConvTranspose2d(
                    128, up_classes, kernel_size=4, stride=2, padding=1, bias=False
                ),
                BatchNorm(up_classes),
                nn.ReLU(inplace=True),
            )

    def forward(self, x):
        out = self.mapping(x)
        prob = self.cls(out)
        up_feat = self.up(out) if self.up else None
        return prob, up_feat


class HD3(BaseModel):
    pretrained_checkpoints = {
        "chairs": "https://github.com/hmorimitsu/ptlflow/releases/download/weights1/hd3-chairs-0d46c9fd.ckpt",
        "things": "https://github.com/hmorimitsu/ptlflow/releases/download/weights1/hd3-things-afcd2eb4.ckpt",
        "sintel": "https://github.com/hmorimitsu/ptlflow/releases/download/weights1/hd3-sintel-10689995.ckpt",
        "kitti": "https://github.com/hmorimitsu/ptlflow/releases/download/weights1/hd3-kitti-6eb77dd3.ckpt",
    }

    def __init__(
        self,
        task: str = "flow",
        encoder: str = "dlaup",
        decoder: str = "hda",
        downsample: int = 6,
        corr_range: Sequence[int] = (4, 4, 4, 4, 4, 4),
        context: bool = False,
        **kwargs,
    ):
        super(HD3, self).__init__(
            loss_fn=LossCalculator(task), output_stride=int(2**downsample), **kwargs
        )

        self.task = task
        self.encoder = encoder
        self.decoder = decoder
        self.downsample = downsample
        self.corr_range = corr_range
        self.context = context

        if self.task == "flow":
            self.corr_range = self.corr_range[:5]

        self.task = self.task
        self.dim = 1 if self.task == "stereo" else 2
        self.levels = len(self.corr_range)
        if self.task == "flow":
            self.classes = [(2 * d + 1) ** 2 for d in self.corr_range]
        else:
            self.classes = [2 * d + 1 for d in self.corr_range]

        if self.encoder == "vgg":
            pyr_channels = [16, 32, 64, 96, 128, 196]
            assert self.levels <= len(pyr_channels)
            self.encoder = vgg_encoder(pyr_channels)
        elif self.encoder == "dlaup":
            pyr_channels = [16, 32, 64, 128, 256, 512, 512]
            self.encoder = dlaup_encoder(pyr_channels)
        else:
            raise ValueError("Unknown encoder {}".format(self.encoder))

        if self.decoder == "resnet":
            dec_block = PreDecoder.ResnetDecoder
        elif self.decoder == "hda":
            dec_block = PreDecoder.HDADecoder
        else:
            raise ValueError("Unknown decoder {}".format(self.decoder))

        feat_d_offset = pyr_channels[::-1]
        feat_d_offset[0] = 0
        up_d_offset = [0] + self.classes[1:]

        for l in range(self.levels):
            setattr(self, "cost_bn_{}".format(l), BatchNorm(self.classes[l]))
            input_d = (
                self.classes[l] + feat_d_offset[l] + up_d_offset[l] + self.dim * (l > 0)
            )
            if l < self.levels - 1:
                up_classes = self.classes[l + 1]
            else:
                up_classes = -1
            if self.context and l == self.levels - 1:
                setattr(self, "Decoder_{}".format(l), Context(input_d, self.classes[l]))
            else:
                setattr(
                    self,
                    "Decoder_{}".format(l),
                    Decoder(input_d, dec_block, self.classes[l], up_classes=up_classes),
                )

        self.corr = SpatialCorrelationSampler(kernel_size=1, patch_size=9, padding=0)

        for m in self.modules():
            classname = m.__class__.__name__
            if isinstance(m, (nn.Conv2d, nn.ConvTranspose2d)):
                m.weight.data.normal_(0, 0.02)
                if m.bias is not None:
                    m.bias.data.zero_()
            elif classname.find("BatchNorm") != -1:
                m.weight.data.fill_(1)
                m.bias.data.zero_()

    def shift(self, x, vect):
        if vect.size(1) < 2:
            vect = disp2flow(vect)
        return flow_warp(x, vect)

    def forward(self, inputs):
        x, image_resizer = self.preprocess_images(
            inputs["images"],
            bgr_add=[-0.485, -0.456, -0.406],
            bgr_mult=[1 / 0.229, 1 / 0.224, 1 / 0.225],
            bgr_to_rgb=False,
            resize_mode="interpolation",
            interpolation_mode="bilinear",
            interpolation_align_corners=True,
        )

        # extract pyramid features
        bs = x.size(0)
        feat_list = self.encoder(torch.cat([x[:, 0], x[:, 1]], 0))
        fp_0 = [f[:bs, :, :, :] for f in feat_list[::-1]]
        fp_1 = [f[bs:, :, :, :] for f in feat_list[::-1]]

        ms_pred = []
        for l in range(self.levels):
            ref_feat = fp_0[l]
            tar_feat = fp_1[l]

            if l == 0:
                tar_feat_corr = tar_feat
            else:
                tar_feat_corr = self.shift(tar_feat, up_curr_vect)

            cost_vol = self.corr(ref_feat, tar_feat_corr)
            cost_vol = cost_vol.view(
                cost_vol.shape[0], -1, cost_vol.shape[3], cost_vol.shape[4]
            )
            cost_vol = cost_vol / ref_feat.shape[1]
            if self.task == "stereo":
                c = self.classes[l] // 2
                cost_vol = cost_vol[
                    :, c * (2 * c + 1) : (c + 1) * (2 * c + 1), :, :
                ].contiguous()
            cost_bn = getattr(self, "cost_bn_" + str(l))
            cost_vol = cost_bn(cost_vol)

            if l == 0:
                decoder_input = cost_vol
            else:
                decoder_input = torch.cat(
                    [cost_vol, ref_feat, ms_pred[-1][-1], up_curr_vect], 1
                )

            decoder = getattr(self, "Decoder_" + str(l))
            prob_map, up_feat = decoder(decoder_input)

            curr_vect = density2vector(prob_map, self.dim, True)
            if l > 0:
                curr_vect += up_curr_vect
            if self.task == "stereo":
                curr_vect = torch.clamp(curr_vect, max=0)
            ms_pred.append([prob_map, curr_vect * 2 ** (self.downsample - l), up_feat])

            if l < self.levels - 1:
                up_curr_vect = 2 * F.interpolate(
                    curr_vect, scale_factor=2, mode="bilinear", align_corners=True
                )

        ms_prob = [l[0] for l in ms_pred]
        ms_vect = [l[1] for l in ms_pred]

        flow_up = F.interpolate(
            ms_vect[-1], scale_factor=4, mode="bilinear", align_corners=False
        )
        flow_up = self.postprocess_predictions(flow_up, image_resizer, is_flow=True)

        outputs = {}
        if self.training:
            outputs["ms_prob"] = ms_prob
            outputs["ms_pred"] = ms_vect
            outputs["corr_range"] = self.corr_range
            outputs["downsample"] = self.downsample
            outputs["flows"] = flow_up[:, None]
        else:
            outputs["flows"] = flow_up[:, None]
        return outputs


class HD3Context(HD3):
    pretrained_checkpoints = {
        "chairs": "https://github.com/hmorimitsu/ptlflow/releases/download/weights1/hd3_ctxt-chairs-d7448468.ckpt",
        "things": "https://github.com/hmorimitsu/ptlflow/releases/download/weights1/hd3_ctxt-things-f1681f75.ckpt",
        "sintel": "https://github.com/hmorimitsu/ptlflow/releases/download/weights1/hd3_ctxt-sintel-22e655e9.ckpt",
        "kitti": "https://github.com/hmorimitsu/ptlflow/releases/download/weights1/hd3_ctxt-kitti-e7d69776.ckpt",
    }

    def __init__(
        self,
        task: str = "flow",
        encoder: str = "dlaup",
        decoder: str = "hda",
        downsample: int = 6,
        corr_range: Sequence[int] = (4, 4, 4, 4, 4, 4),
        context: bool = True,
        **kwargs,
    ):
        super().__init__(
            task, encoder, decoder, downsample, corr_range, context, **kwargs
        )


@register_model
@trainable
class hd3(HD3):
    pass


@register_model
@trainable
class hd3_ctxt(HD3Context):
    pass



import torch
import torch.nn.functional as F


def normalize_img(img, mean, std):
    return (img / 255.0 - mean) / std


def coords_grid(b, h, w, dtype, device):
    ys, xs = torch.meshgrid(
        torch.arange(h, dtype=dtype, device=device),
        torch.arange(w, dtype=dtype, device=device),
        indexing="ij",
    )  # [H, W]

    stacks = [xs, ys]

    grid = torch.stack(stacks, dim=0)  # [2, H, W] or [3, H, W]

    grid = grid[None].repeat(b, 1, 1, 1)  # [B, 2, H, W] or [B, 3, H, W]

    return grid


def bilinear_sample(img, sample_coords):
    b, _, h, w = sample_coords.shape

    # Normalize to [-1, 1]
    x_grid = 2 * sample_coords[:, 0] / (w - 1) - 1
    y_grid = 2 * sample_coords[:, 1] / (h - 1) - 1

    grid = torch.stack([x_grid, y_grid], dim=-1)  # [B, H, W, 2]

    img = F.grid_sample(
        img, grid, mode="bilinear", padding_mode="zeros", align_corners=True
    )

    return img


def flow_warp(feature, flow):
    b, c, h, w = feature.size()

    grid = (
        coords_grid(b, h, w, dtype=flow.dtype, device=flow.device) + flow
    )  # [B, 2, H, W]

    return bilinear_sample(feature, grid)



# =============================================================================
# Copyright 2023 Henrique Morimitsu
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# =============================================================================

import math
from typing import Optional

import torch
import torch.nn as nn
import torch.nn.functional as F


def pklinear(
    x: torch.Tensor,
    weight: torch.Tensor,
    bias=None,
    pre_act_fn=None,
    post_act_fn=None,
    out_ch: Optional[int] = None,
) -> torch.Tensor:
    if out_ch is None:
        out_ch = weight.shape[0]

    if pre_act_fn is not None:
        x = pre_act_fn(x)

    w = weight[:out_ch, : x.shape[-1]]
    b = None
    if bias is not None:
        b = bias[:out_ch]
    x = F.linear(x, weight=w, bias=b)

    if post_act_fn is not None:
        x = post_act_fn(x)

    return x


class PKLinear(nn.Module):
    def __init__(
        self,
        in_features,
        out_features,
        bias=True,
        device=None,
        dtype=None,
        pre_act_fn=None,
        post_act_fn=None,
    ) -> None:
        super().__init__()
        self.in_features = in_features
        self.out_features = out_features
        self.device = device
        self.dtype = dtype
        self.pre_act_fn = pre_act_fn
        self.post_act_fn = post_act_fn

        self.register_parameter(
            "weight",
            nn.Parameter(
                torch.zeros(out_features, in_features, device=device, dtype=dtype)
            ),
        )

        if bias:
            self.register_parameter(
                "bias",
                nn.Parameter(torch.zeros(out_features, device=device, dtype=dtype)),
            )
        else:
            self.bias = None

        self.reset_parameters()

    def reset_parameters(self) -> None:
        # Setting a=sqrt(5) in kaiming_uniform is the same as initializing with
        # uniform(-1/sqrt(in_features), 1/sqrt(in_features)). For details, see
        # https://github.com/pytorch/pytorch/issues/57109
        nn.init.kaiming_uniform_(self.weight, a=math.sqrt(5))
        if self.bias is not None:
            fan_in, _ = nn.init._calculate_fan_in_and_fan_out(self.weight)
            bound = 1 / math.sqrt(fan_in) if fan_in > 0 else 0
            nn.init.uniform_(self.bias, -bound, bound)

    def forward(self, x: torch.Tensor, out_ch: Optional[int] = None) -> torch.Tensor:
        return pklinear(
            x=x,
            weight=self.weight,
            bias=self.bias,
            pre_act_fn=self.pre_act_fn,
            post_act_fn=self.post_act_fn,
            out_ch=out_ch,
        )



import math
from typing import Sequence

import torch
import torch.nn as nn
import torch.nn.functional as F

from ptlflow.utils.registry import register_model, trainable
from .update import BasicUpdateBlock
from .corr import get_corr_block
from .utils import coords_grid
from .extractor import ResNetFPN
from .layer import conv3x3
from ..base_model.base_model import BaseModel

try:
    import alt_cuda_corr
except:
    alt_cuda_corr = None


class SequenceLoss(nn.Module):
    def __init__(self, gamma: float, max_flow: float):
        super().__init__()
        self.gamma = gamma
        self.max_flow = max_flow

    def forward(self, outputs, inputs):
        """Loss function defined over sequence of flow predictions"""
        flow_preds = outputs["flow_preds"]
        flow_gt = inputs["flows"][:, 0]
        valid = inputs["valids"][:, 0]

        n_predictions = len(flow_preds)

        flow_loss = 0.0
        # exlude invalid pixels and extremely large diplacements
        mag = torch.sum(flow_gt**2, dim=1, keepdim=True).sqrt()
        valid = (valid >= 0.5) & (mag < self.max_flow)
        for i in range(n_predictions):
            i_weight = self.gamma ** (n_predictions - i - 1)
            loss_i = outputs["nf_preds"][i]
            final_mask = (
                (~torch.isnan(loss_i.detach()))
                & (~torch.isinf(loss_i.detach()))
                & valid
            )
            flow_loss += i_weight * ((final_mask * loss_i).sum() / final_mask.sum())

        return flow_loss


class SEARAFT(BaseModel):
    def __init__(
        self,
        corr_levels: int = 4,
        corr_radius: int = 4,
        dim: int = 128,
        initial_dim: int = 64,
        num_blocks: int = 2,
        block_dims: Sequence[int] = (64, 128, 256),
        pretrain: str = "resnet18",
        gamma: float = 0.8,
        max_flow: float = 400,
        iters: int = 4,
        alternate_corr: bool = False,
        use_var: bool = True,
        var_min: float = 0,
        var_max: float = 10,
        **kwargs,
    ) -> None:
        super().__init__(
            output_stride=8, loss_fn=SequenceLoss(gamma, max_flow), **kwargs
        )

        self.corr_levels = corr_levels
        self.corr_radius = corr_radius
        self.dim = dim
        self.initial_dim = initial_dim
        self.num_blocks = num_blocks
        self.block_dims = block_dims
        self.pretrain = pretrain
        self.gamma = gamma
        self.max_flow = max_flow
        self.iters = iters
        self.alternate_corr = alternate_corr
        self.use_var = use_var
        self.var_min = var_min
        self.var_max = var_max

        self.output_dim = dim * 2

        self.corr_channel = corr_levels * (corr_radius * 2 + 1) ** 2
        self.cnet = ResNetFPN(
            block_dims=block_dims,
            initial_dim=initial_dim,
            pretrain=pretrain,
            input_dim=6,
            output_dim=2 * self.dim,
            norm_layer=nn.BatchNorm2d,
            init_weight=True,
        )

        # conv for iter 0 results
        self.init_conv = conv3x3(2 * dim, 2 * dim)
        self.upsample_weight = nn.Sequential(
            # convex combination of 3x3 patches
            nn.Conv2d(dim, dim * 2, 3, padding=1),
            nn.ReLU(inplace=True),
            nn.Conv2d(dim * 2, 64 * 9, 1, padding=0),
        )
        self.flow_head = nn.Sequential(
            # flow(2) + weight(2) + log_b(2)
            nn.Conv2d(dim, 2 * dim, 3, padding=1),
            nn.ReLU(inplace=True),
            nn.Conv2d(2 * dim, 6, 3, padding=1),
        )
        if iters > 0:
            self.fnet = ResNetFPN(
                block_dims=block_dims,
                initial_dim=initial_dim,
                pretrain=pretrain,
                input_dim=3,
                output_dim=self.output_dim,
                norm_layer=nn.BatchNorm2d,
                init_weight=True,
            )
            self.update_block = BasicUpdateBlock(
                corr_channel=self.corr_channel,
                num_blocks=num_blocks,
                hdim=dim,
                cdim=dim,
            )

        if self.alternate_corr and alt_cuda_corr is None:
            print(
                "!!! alt_cuda_corr is not compiled! The slower IterativeCorrBlock will be used instead !!!"
            )

    def initialize_flow(self, img):
        """Flow is represented as difference between two coordinate grids flow = coords2 - coords1"""
        N, C, H, W = img.shape
        coords1 = coords_grid(N, H // 8, W // 8, dtype=img.dtype, device=img.device)
        coords2 = coords_grid(N, H // 8, W // 8, dtype=img.dtype, device=img.device)
        return coords1, coords2

    def upsample_data(self, flow, info, mask):
        """Upsample [H/8, W/8, C] -> [H, W, C] using convex combination"""
        N, C, H, W = info.shape
        mask = mask.view(N, 1, 9, 8, 8, H, W)
        mask = torch.softmax(mask, dim=2)

        up_flow = F.unfold(8 * flow, [3, 3], padding=1)
        up_flow = up_flow.view(N, 2, 9, 1, 1, H, W)
        up_info = F.unfold(info, [3, 3], padding=1)
        up_info = up_info.view(N, C, 9, 1, 1, H, W)

        up_flow = torch.sum(mask * up_flow, dim=2)
        up_flow = up_flow.permute(0, 1, 4, 2, 5, 3)
        up_info = torch.sum(mask * up_info, dim=2)
        up_info = up_info.permute(0, 1, 4, 2, 5, 3)

        return up_flow.reshape(N, 2, 8 * H, 8 * W), up_info.reshape(N, C, 8 * H, 8 * W)

    def forward(self, inputs):
        """Estimate optical flow between pair of frames"""
        images, image_resizer = self.preprocess_images(
            inputs["images"],
            bgr_add=-0.5,
            bgr_mult=2.0,
            bgr_to_rgb=True,
            resize_mode="pad",
            pad_mode="replicate",
            pad_two_side=True,
        )

        image1 = images[:, 0]
        image2 = images[:, 1]

        if "flows" in inputs:
            flow_gt = inputs["flows"][:, 0]
        else:
            N, _, H, W = image1.shape
            flow_gt = torch.zeros(N, 2, H, W, device=image1.device)

        flow_predictions = []
        info_predictions = []

        # run the context network
        cnet = self.cnet(torch.cat([image1, image2], dim=1))
        cnet = self.init_conv(cnet)
        net, context = torch.split(cnet, [self.dim, self.dim], dim=1)

        # init flow
        flow_update = self.flow_head(net)
        weight_update = 0.25 * self.upsample_weight(net)
        flow_8x = flow_update[:, :2]
        info_8x = flow_update[:, 2:]
        flow_up, info_up = self.upsample_data(flow_8x, info_8x, weight_update)
        flow_predictions.append(flow_up)
        info_predictions.append(info_up)

        if self.iters > 0:
            # run the feature network
            fmap1_8x = self.fnet(image1)
            fmap2_8x = self.fnet(image2)
            corr_fn = get_corr_block(
                fmap1=fmap1_8x,
                fmap2=fmap2_8x,
                radius=self.corr_radius,
                num_levels=self.corr_levels,
                alternate_corr=self.alternate_corr,
            )

        for itr in range(self.iters):
            N, _, H, W = flow_8x.shape
            flow_8x = flow_8x.detach()
            coords2 = (
                coords_grid(N, H, W, dtype=image1.dtype, device=image1.device) + flow_8x
            ).detach()
            corr = corr_fn(coords2)
            net = self.update_block(net, context, corr, flow_8x)
            flow_update = self.flow_head(net)
            weight_update = 0.25 * self.upsample_weight(net)
            flow_8x = flow_8x + flow_update[:, :2]
            info_8x = flow_update[:, 2:]
            # upsample predictions
            flow_up, info_up = self.upsample_data(flow_8x, info_8x, weight_update)
            flow_up = self.postprocess_predictions(flow_up, image_resizer, is_flow=True)
            info_up = self.postprocess_predictions(
                info_up, image_resizer, is_flow=False
            )
            flow_predictions.append(flow_up)
            info_predictions.append(info_up)

        if self.training:
            # exlude invalid pixels and extremely large diplacements
            nf_predictions = []
            for i in range(len(info_predictions)):
                if not self.use_var:
                    var_max = var_min = 0
                else:
                    var_max = self.var_max
                    var_min = self.var_min

                raw_b = info_predictions[i][:, 2:]
                log_b = torch.zeros_like(raw_b)
                weight = info_predictions[i][:, :2]
                # Large b Component
                log_b[:, 0] = torch.clamp(raw_b[:, 0], min=0, max=var_max)
                # Small b Component
                log_b[:, 1] = torch.clamp(raw_b[:, 1], min=var_min, max=0)
                # term2: [N, 2, m, H, W]
                term2 = ((flow_gt - flow_predictions[i]).abs().unsqueeze(2)) * (
                    torch.exp(-log_b).unsqueeze(1)
                )
                # term1: [N, m, H, W]
                term1 = weight - math.log(2) - log_b
                nf_loss = torch.logsumexp(
                    weight, dim=1, keepdim=True
                ) - torch.logsumexp(term1.unsqueeze(1) - term2, dim=2)
                nf_predictions.append(nf_loss)

        if self.training:
            outputs = {
                "flows": flow_up[:, None],
                "flow_preds": flow_predictions,
                "info_preds": info_predictions,
                "nf_preds": nf_predictions,
            }
        else:
            outputs = {"flows": flow_up[:, None], "flow_small": flow_8x}

        return outputs


class SEARAFT_S(SEARAFT):
    pretrained_checkpoints = {
        "tartan": "https://github.com/hmorimitsu/ptlflow/releases/download/weights1/sea_raft_s-tartan-f7e26f21.ckpt",
        "chairs": "https://github.com/hmorimitsu/ptlflow/releases/download/weights1/sea_raft_s-chairs-6980249f.ckpt",
        "things": "https://github.com/hmorimitsu/ptlflow/releases/download/weights1/sea_raft_s-things-a15c1713.ckpt",
        "sintel": "https://github.com/hmorimitsu/ptlflow/releases/download/weights1/sea_raft_s-sintel-bb63371a.ckpt",
        "kitti": "https://github.com/hmorimitsu/ptlflow/releases/download/weights1/sea_raft_s-kitti-3a96c1cc.ckpt",
        "spring": "https://github.com/hmorimitsu/ptlflow/releases/download/weights1/sea_raft_s-spring-4d13c106.ckpt",
    }

    def __init__(
        self,
        corr_levels: int = 4,
        corr_radius: int = 4,
        dim: int = 128,
        initial_dim: int = 64,
        num_blocks: int = 2,
        block_dims: Sequence[int] = (64, 128, 256),
        pretrain: str = "resnet18",
        gamma: float = 0.8,
        max_flow: float = 400,
        iters: int = 4,
        alternate_corr: bool = False,
        use_var: bool = True,
        var_min: float = 0,
        var_max: float = 10,
        **kwargs,
    ) -> None:
        super().__init__(
            corr_levels,
            corr_radius,
            dim,
            initial_dim,
            num_blocks,
            block_dims,
            pretrain,
            gamma,
            max_flow,
            iters,
            alternate_corr,
            use_var,
            var_min,
            var_max,
            **kwargs,
        )


class SEARAFT_M(SEARAFT):
    pretrained_checkpoints = {
        "tartan": "https://github.com/hmorimitsu/ptlflow/releases/download/weights1/sea_raft_m-tartan-e684ed5f.ckpt",
        "chairs": "https://github.com/hmorimitsu/ptlflow/releases/download/weights1/sea_raft_m-chairs-1cb7b11e.ckpt",
        "things": "https://github.com/hmorimitsu/ptlflow/releases/download/weights1/sea_raft_m-things-ac45dd7f.ckpt",
        "sintel": "https://github.com/hmorimitsu/ptlflow/releases/download/weights1/sea_raft_m-sintel-f8bb7e3f.ckpt",
        "kitti": "https://github.com/hmorimitsu/ptlflow/releases/download/weights1/sea_raft_m-kitti-e51f7603.ckpt",
        "spring": "https://github.com/hmorimitsu/ptlflow/releases/download/weights1/sea_raft_m-spring-de7c13e2.ckpt",
    }

    def __init__(
        self,
        corr_levels: int = 4,
        corr_radius: int = 4,
        dim: int = 128,
        initial_dim: int = 64,
        num_blocks: int = 2,
        block_dims: Sequence[int] = (64, 128, 256),
        pretrain: str = "resnet34",
        gamma: float = 0.8,
        max_flow: float = 400,
        iters: int = 4,
        alternate_corr: bool = False,
        use_var: bool = True,
        var_min: float = 0,
        var_max: float = 10,
        **kwargs,
    ) -> None:
        super().__init__(
            corr_levels,
            corr_radius,
            dim,
            initial_dim,
            num_blocks,
            block_dims,
            pretrain,
            gamma,
            max_flow,
            iters,
            alternate_corr,
            use_var,
            var_min,
            var_max,
            **kwargs,
        )


class SEARAFT_L(SEARAFT):
    pretrained_checkpoints = {
        "tartan": "https://github.com/hmorimitsu/ptlflow/releases/download/weights1/sea_raft_m-tartan-e684ed5f.ckpt",
        "chairs": "https://github.com/hmorimitsu/ptlflow/releases/download/weights1/sea_raft_m-chairs-1cb7b11e.ckpt",
        "things": "https://github.com/hmorimitsu/ptlflow/releases/download/weights1/sea_raft_m-things-ac45dd7f.ckpt",
        "sintel": "https://github.com/hmorimitsu/ptlflow/releases/download/weights1/sea_raft_m-sintel-f8bb7e3f.ckpt",
        "kitti": "https://github.com/hmorimitsu/ptlflow/releases/download/weights1/sea_raft_m-kitti-e51f7603.ckpt",
        "spring": "https://github.com/hmorimitsu/ptlflow/releases/download/weights1/sea_raft_m-spring-de7c13e2.ckpt",
    }

    def __init__(
        self,
        corr_levels: int = 4,
        corr_radius: int = 4,
        dim: int = 128,
        initial_dim: int = 64,
        num_blocks: int = 2,
        block_dims: Sequence[int] = (64, 128, 256),
        pretrain: str = "resnet34",
        gamma: float = 0.8,
        max_flow: float = 400,
        iters: int = 12,
        alternate_corr: bool = False,
        use_var: bool = True,
        var_min: float = 0,
        var_max: float = 10,
        **kwargs,
    ) -> None:
        super().__init__(
            corr_levels,
            corr_radius,
            dim,
            initial_dim,
            num_blocks,
            block_dims,
            pretrain,
            gamma,
            max_flow,
            iters,
            alternate_corr,
            use_var,
            var_min,
            var_max,
            **kwargs,
        )


@register_model
@trainable
class sea_raft(SEARAFT):
    pass


@register_model
@trainable
class sea_raft_s(SEARAFT_S):
    pass


@register_model
@trainable
class sea_raft_m(SEARAFT_M):
    pass


@register_model
@trainable
class sea_raft_l(SEARAFT_L):
    pass



""" Normalization layers and wrappers

Norm layer definitions that support fast norm and consistent channel arg order (always first arg).

Hacked together by / Copyright 2022 Ross Wightman
"""

import torch
import torch.nn as nn
import torch.nn.functional as F


class LayerNorm2d(nn.LayerNorm):
    """LayerNorm for channels of '2D' spatial NCHW tensors"""

    def __init__(self, num_channels, eps=1e-6, affine=True):
        super().__init__(num_channels, eps=eps, elementwise_affine=affine)

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        x = x.permute(0, 2, 3, 1)
        if self.weight is not None:
            x = F.layer_norm(
                x,
                (x.shape[-1],),
                self.weight,
                self.bias,
                self.eps,
            )
        else:
            x = F.layer_norm(x, (x.shape[-1],), eps=self.eps)
        x = x.permute(0, 3, 1, 2)
        return x


class GroupNorm2d(nn.GroupNorm):
    """GroupNorm for channels of '2D' spatial NCHW tensors"""

    def __init__(self, num_groups, num_channels, eps=1e-6, affine=True):
        super().__init__(num_groups, num_channels, eps=eps, affine=affine)

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        if self.weight is not None:
            x = F.group_norm(
                x,
                self.num_groups,
                self.weight,
                self.bias,
                self.eps,
            )
        else:
            x = F.group_norm(x, self.num_groups, eps=self.eps)
        return x


class BatchNorm2d(nn.BatchNorm2d):
    def __init__(
        self,
        num_channels: int,
        eps: float = 0.00001,
        momentum: float = 0.1,
        affine: bool = True,
        track_running_stats: bool = True,
        device=None,
        dtype=None,
    ) -> None:
        super().__init__(
            num_channels, eps, momentum, affine, track_running_stats, device, dtype
        )

    def forward(self, input: torch.Tensor) -> torch.Tensor:
        self._check_input_dim(input)

        # exponential_average_factor is set to self.momentum
        # (when it is available) only so that it gets updated
        # in ONNX graph when this node is exported to ONNX.
        if self.momentum is None:
            exponential_average_factor = 0.0
        else:
            exponential_average_factor = self.momentum

        if self.training and self.track_running_stats:
            # TODO: if statement only here to tell the jit to skip emitting this when it is None
            if self.num_batches_tracked is not None:  # type: ignore[has-type]
                self.num_batches_tracked.add_(1)  # type: ignore[has-type]
                if self.momentum is None:  # use cumulative moving average
                    exponential_average_factor = 1.0 / float(self.num_batches_tracked)
                else:  # use exponential moving average
                    exponential_average_factor = self.momentum

        r"""
        Decide whether the mini-batch stats should be used for normalization rather than the buffers.
        Mini-batch stats are used in training mode, and in eval mode when buffers are None.
        """
        if self.training:
            bn_training = True
        else:
            bn_training = (self.running_mean is None) and (self.running_var is None)

        r"""
        Buffers are only updated if they are to be tracked and we are in training mode. Thus they only need to be
        passed when the update should occur (i.e. in training mode when they are tracked), or when buffer stats are
        used for normalization (i.e. in eval mode when buffers are not None).
        """
        return F.batch_norm(
            input,
            # If buffers are not to be tracked, ensure that they won't be updated
            (
                self.running_mean[: input.shape[1]]
                if not self.training or self.track_running_stats
                else None
            ),
            (
                self.running_var[: input.shape[1]]
                if not self.training or self.track_running_stats
                else None
            ),
            self.weight,
            self.bias,
            bn_training,
            exponential_average_factor,
            self.eps,
        )
